from math import cos, pi, sin

from glm import lookAt, vec3

from lsystem.graphics.spatial_object import SpatialObject


class OrbitalObject(SpatialObject):
    def __init__(self):
        super().__init__()
        self.r = 2
        self.theta = 0
        self.psi = 0
        self.deg = pi / 180.0
        self.origin = vec3(0)

    def set_position(self, r, theta, psi):
        self.r = r
        self.theta = theta
        self.psi = psi
        self.position = vec3(
            r * cos(psi * self.deg) * cos(theta * self.deg),
            r * sin(psi * self.deg),
            r * cos(psi * self.deg) * sin(theta * self.deg),
        )

    def get_position(self):
        self.position = vec3(
            self.r * cos(self.psi * self.deg) * cos(self.theta * self.deg),
            self.r * sin(self.psi * self.deg),
            self.r * cos(self.psi * self.deg) * sin(self.theta * self.deg),
        )
        return self.position

    def set_origin(self, origin):
        self.origin = vec3(origin)

    def add_theta(self, num):
        self.theta += num
        if self.theta > 360:
            self.theta -= 360
        if self.theta < 0:
            self.theta += 360
        self.set_position(self.r, self.theta, self.psi)

    def add_r(self, num):
        self.r += num
        if self.r < 0.000001:
            self.r = 0.000001
        self.set_position(self.r, self.theta, self.psi)

    def add_psi(self, num):
        self.psi += num
        if self.psi > 90:
            self.psi = 90
        if self.psi < -90:
            self.psi = -90
        self.set_position(self.r, self.theta, self.psi)

    def set_psi(self, num):
        self.psi = num
        if self.psi > 90:
            self.psi = 90
        if self.psi < -90:
            self.psi = -90
        self.set_position(self.r, self.theta, self.psi)

    def set_r(self, num):
        self.r = num
        if self.r < 0.000001:
            self.r = 0.000001
        self.set_position(self.r, self.theta, self.psi)

    def set_theta(self, num):
        self.num = num
        while self.theta > 360:
            self.theta -= 360
        while self.theta < 0:
            self.theta += 360
        self.set_position(self.r, self.theta, self.psi)

    def get_theta(self):
        return self.theta

    def get_r(self):
        return self.r

    def get_psi(self):
        return self.psi

    def get_facing(self):
        eye = vec3(
            self.r * cos(self.psi * self.deg) * cos(self.theta * self.deg),
            self.r * sin(self.psi * self.deg),
            self.r * cos(self.psi * self.deg) * sin(self.theta * self.deg),
        )
        center = self.origin
        up = vec3(0, 1, 0)
        return lookAt(eye, center, up)



import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection, Line3DCollection
from pyrr import line, ray, geometric_tests, plane
from math import radians
from multiprocessing import Process, Lock, cpu_count, Array, Value, Queue, connection, Pool
import math
import ctypes as c


class Cubes:
    """This is a utility class to create a grid of cubes around a 3D mesh for the purposes of calculating the fractal dimension.
    Attributes:
        n (int): number of sides on each face.
        r (float): Controls the radius of the cubes. Don't touch this.
        cube_array (array): Shared memory array for multiprocessing.
        cubes (array): Array of cubes tracking whether we've intersected them.
        xy_rays (array): Array of vectors used for raycasting the mesh vertices into faces of the cubes.
        xz_rays (array): Array of vectors used for raycasting the mesh vertices into faces of the cubes.
        yz_rays (array): Array of vectors used for raycasting the mesh vertices into faces of the cubes.
        max_t (int): Maximum number of supported threads on the system.
        available_threads: Current number of available threads.
    """
    def __init__(self, n):
        """Constructor for the Cubes class.
        Parameters:
            n (int): Number of faces on each side of the cubic grid.
        """
        self.n = n
        self.r = 0.5
        self.cube_array = Array(c.c_double, n * n * n)
        self.cubes = np.frombuffer(self.cube_array.get_obj())
        self.cubes.fill(0)
        self.cubes = self.cubes.reshape((n, n, n))
        self.xy_rays = []
        self.xz_rays = []
        self.yz_rays = []
        # self.max_t = cpu_count()  # total threads
        self.max_t = 24
        self.available_threads = Value('i', self.max_t)  # current thread count
        print("Detected " + str(self.max_t) + " threads")
        print("Generating %d rays on %d threads." % (3 * (n + 1) ** 2, self.max_t))
        self.generate_xy_rays()
        self.generate_xz_rays()
        self.generate_yz_rays()

    def process_queue(self, queue, rays):
        while(not queue.empty()):
            val = queue.get()
            rays[val[0]] += val[1]


    def sub_generate_xz_rays(self, x, queue):
        """Subprocess for generating xz rays."""
        rays = []
        for z in range(self.n + 1):
            rays += [
                ray.create_from_line(
                    line.create_from_points((x, 0, z), (x, self.n, z))
                )
            ]
        queue.put((x, rays))
        queue.close()

    def sub_generate_xy_rays(self, x, queue):
        """Subprocess for generating xy rays."""
        rays = []
        for y in range(self.n + 1):
            rays += [
                ray.create_from_line(
                    line.create_from_points((x, y, 0), (x, y, self.n))
                )
            ]
        queue.put((x, rays))
        queue.close()

    def sub_generate_yz_rays(self, y, queue):
        """Subprocess for generating yz rays."""
        rays = []
        for z in range(self.n + 1):
            rays += [
                ray.create_from_line(
                    line.create_from_points((0, y, z), (self.n, y, z))
                )
            ]
        queue.put((y, rays))
        queue.close()

    def generate_xy_rays(self):
        """Populates the xy_rays array with raycasting vectors on the xy plane from 0 to n"""
        print("Generating xy")
        queue = Queue()
        processes = []
        for x in range(self.n + 1):
            self.xy_rays += [[]]
            if(self.available_threads.value >= 1):
                processes.append(
                    Process(target=self.sub_generate_xy_rays, args=(x, queue))
                    )
                processes[-1].start()
                self.available_threads.value -= 1
            else:
                while True:
                    self.process_queue(queue, self.xy_rays)
                    if (processes[0].is_alive() is not True):
                        processes[0].join(timeout=1)
                        break
                processes.pop(0)
                self.available_threads.value += 1
                processes.append(
                    Process(target=self.sub_generate_xy_rays, args=(x, queue))
                    )
                processes[-1].start()
                self.available_threads.value -= 1

        for p in processes:
            while True:
                self.process_queue(queue, self.xy_rays)
                if (p.is_alive() is not True):
                    p.join(timeout=1)
                    break
            self.available_threads.value += 1
        self.process_queue(queue, self.xy_rays)

    def generate_xz_rays(self):
        """Populates the xz_rays array with raycasting vectors on the xz plane from 0 to n"""
        print("Generating xz")
        queue = Queue()
        processes = []
        for x in range(self.n + 1):
            self.xz_rays += [[]]
            if(self.available_threads.value >= 1):
                processes.append(
                    Process(target=self.sub_generate_xz_rays, args=(x, queue))
                    )
                processes[-1].start()
                self.available_threads.value -= 1
            else:
                while True:
                    self.process_queue(queue, self.xz_rays)
                    if (processes[0].is_alive() is not True):
                        processes[0].join(timeout=1)
                        break
                processes.pop(0)
                self.available_threads.value += 1
                processes.append(
                    Process(target=self.sub_generate_xz_rays, args=(x, queue))
                    )
                processes[-1].start()
                self.available_threads.value -= 1
        for p in processes:
            while True:
                self.process_queue(queue, self.xz_rays)
                if (p.is_alive() is not True):
                    p.join(timeout=1)
                    break
            self.available_threads.value += 1
        self.process_queue(queue, self.xz_rays)


    def generate_yz_rays(self):
        """Populates the yz_rays array with raycasting vectors on the yz plane from 0 to n"""
        print("Generating yz")
        queue = Queue()
        processes = []
        for y in range(self.n + 1):
            self.yz_rays += [[]]
            if(self.available_threads.value >= 1):
                processes.append(
                    Process(target=self.sub_generate_yz_rays, args=(y, queue))
                    )
                processes[-1].start()
                self.available_threads.value -= 1
            else:
                while True:
                    self.process_queue(queue, self.yz_rays)
                    if (processes[0].is_alive() is not True):
                        processes[0].join(timeout=1)
                        break
                processes.pop(0)
                self.available_threads.value += 1
                processes.append(
                    Process(target=self.sub_generate_yz_rays, args=(y, queue))
                    )
                processes[-1].start()
                self.available_threads.value -= 1

        for p in processes:
            while True:
                self.process_queue(queue, self.yz_rays)
                if (p.is_alive() is not True):
                    p.join(timeout=1)
                    break
            p.join()
            self.available_threads.value += 1
        self.process_queue(queue, self.yz_rays)

    def check_bounds(self, verts, isec):
        """Validates the bounds of the intersection with the min/max values of the min/max values of the mesh."""
        min_v = verts[0]
        max_v = verts[1]
        isec = np.array(isec)
        x = (isec[0] >= min_v[0]) & (isec[0] <= max_v[0])
        y = (isec[1] >= min_v[1]) & (isec[1] <= max_v[1])
        z = (isec[2] >= min_v[2]) & (isec[2] <= max_v[2])
        return x & y & z

    def check_intersection_xy(self, isec, xy, verts):
        """Checks for intersections of a specifiy ray on the xy plane for the mesh"""
        if isec is not None:
            x = xy[0]
            y = xy[1]
            for z in range(self.n):
                if self.cubes[x][y][z] == 0 and isec[2] >= z and isec[2] <= z + 1:
                    if self.check_bounds(verts, isec):
                        self.cubes[x][y][z] = 1
                        return 1
        return 0

    def check_intersection_xz(self, isec, xz, verts):
        """Checks for intersections of a specifiy ray on the xz plane for the mesh"""
        if isec is not None:
            x = xz[0]
            z = xz[1]
            for y in range(self.n):
                if self.cubes[x][y][z] == 0 and isec[1] >= y and isec[1] <= y + 1:
                    if self.check_bounds(verts, isec):
                        self.cubes[x][y][z] = 1
                        return 1
        return 0

    def check_intersection_yz(self, isec, yz, verts):
        """Checks for intersections of a specifiy ray on the yz plane for the mesh"""
        if isec is not None:
            y = yz[0]
            z = yz[1]
            for x in range(self.n):
                if self.cubes[x][y][z] == 0 and isec[0] >= x and isec[0] <= x + 1:
                    if self.check_bounds(verts, isec):
                        self.cubes[x][y][z] = 1
                        return 1
        return 0

    def sub_intersect_xz(self, x, p, min_verts, max_verts):
        """Subprocess for calculating intersections on the xz plane."""
        i = 0
        for z in range(self.n):
            if np.min(self.cubes[x, :, z]) == 0.0:
                bl = geometric_tests.ray_intersect_plane(self.xz_rays[x][z], p)
                br = geometric_tests.ray_intersect_plane(self.xz_rays[x + 1][z], p)
                tl = geometric_tests.ray_intersect_plane(self.xz_rays[x][z + 1], p)
                tr = geometric_tests.ray_intersect_plane(
                    self.xz_rays[x + 1][z + 1], p
                )
                i += self.check_intersection_xz(
                    bl, (x, z), (min_verts, max_verts)
                )
                i += self.check_intersection_xz(
                    br, (x, z), (min_verts, max_verts)
                )
                i += self.check_intersection_xz(
                    tl, (x, z), (min_verts, max_verts)
                )
                i += self.check_intersection_xz(
                    tr, (x, z), (min_verts, max_verts)
                )
        return i

    def intersect_xz(self, verts):
        """This method checks for intersections with the xz vectors. It marks the cubes as intersected and returns the number of intersections."""
        p = plane.create_from_points(verts[0], verts[1], verts[2])
        max_verts = (np.max(verts[:, 0]), np.max(verts[:, 1]), np.max(verts[:, 2]))
        min_verts = (np.min(verts[:, 0]), np.min(verts[:, 1]), np.min(verts[:, 2]))
        inters = 0
        for x in range(self.n):
          inters += self.sub_intersect_xz(x, p, min_verts, max_verts)
        return inters

    def sub_intersect_xy(self, x, p, min_verts, max_verts):
        """Subprocess for calculating intersections on the xy plane."""
        i = 0
        for y in range(self.n):
            if np.min(self.cubes[x, y, :]) == 0.0:
                bl = geometric_tests.ray_intersect_plane(self.xy_rays[x][y], p)
                br = geometric_tests.ray_intersect_plane(self.xy_rays[x + 1][y], p)
                tl = geometric_tests.ray_intersect_plane(self.xy_rays[x][y + 1], p)
                tr = geometric_tests.ray_intersect_plane(
                    self.xy_rays[x + 1][y + 1], p
                )
                i += self.check_intersection_xy(
                    bl, (x, y), (min_verts, max_verts)
                )
                i += self.check_intersection_xy(
                    br, (x, y), (min_verts, max_verts)
                )
                i += self.check_intersection_xy(
                    tl, (x, y), (min_verts, max_verts)
                )
                i += self.check_intersection_xy(
                    tr, (x, y), (min_verts, max_verts)
                )
        return i


    def intersect_xy(self, verts):
        """This method checks for intersections with the xz vectors. It marks the cubes as intersected and returns the number of intersections."""
        p = plane.create_from_points(verts[0], verts[1], verts[2])
        max_verts = (np.max(verts[:, 0]), np.max(verts[:, 1]), np.max(verts[:, 2]))
        min_verts = (np.min(verts[:, 0]), np.min(verts[:, 1]), np.min(verts[:, 2]))
        inters = 0
        for x in range(self.n):
          inters += self.sub_intersect_xy(x, p, min_verts, max_verts)
        return inters

    def sub_intersect_yz(self, y, p, min_verts, max_verts):
        i = 0
        for z in range(self.n):
            if np.min(self.cubes[:, y, z]) == 0.0:
                bl = geometric_tests.ray_intersect_plane(self.yz_rays[y][z], p)
                br = geometric_tests.ray_intersect_plane(self.yz_rays[y + 1][z], p)
                tl = geometric_tests.ray_intersect_plane(self.yz_rays[y][z + 1], p)
                tr = geometric_tests.ray_intersect_plane(
                    self.yz_rays[y + 1][z + 1], p
                )
                i += self.check_intersection_yz(
                    bl, (y, z), (min_verts, max_verts)
                )
                i += self.check_intersection_yz(
                    br, (y, z), (min_verts, max_verts)
                )
                i += self.check_intersection_yz(
                    tl, (y, z), (min_verts, max_verts)
                )
                i += self.check_intersection_yz(
                    tr, (y, z), (min_verts, max_verts)
                )
        return i

    def intersect_yz(self, verts):
        """This method checks for intersections with the yz vectors. It marks the cubes as intersected and returns the number of intersections."""
        p = plane.create_from_points(verts[0], verts[1], verts[2])
        max_verts = (np.max(verts[:, 0]), np.max(verts[:, 1]), np.max(verts[:, 2]))
        min_verts = (np.min(verts[:, 0]), np.min(verts[:, 1]), np.min(verts[:, 2]))
        inters = 0
        for y in range(self.n):
          inters += self.sub_intersect_yz(y, p, min_verts, max_verts)
        return inters

    def generate_grid(self):
        """Generates cubes to be drawn in pyplot"""
        cr = self.r
        cube_verts = []  # List of all cube vert arrays
        cube_faces = []  # List of cube faces
        colors = []
        for x in range(self.n):
            for y in range(self.n):
                for z in range(self.n):
                    cwo = np.array((x + cr, y + cr, z + cr))
                    cv = np.array(
                        [
                            cwo - cr,
                            cwo - np.array((-cr, cr, cr)),
                            cwo - np.array((-cr, -cr, cr)),
                            cwo - np.array((cr, -cr, cr)),
                            cwo - np.array((cr, cr, -cr)),
                            cwo - np.array((-cr, cr, -cr)),
                            cwo - np.array((-cr, -cr, -cr)),
                            cwo - np.array((cr, -cr, -cr,)),
                        ]
                    )
                    cube_verts += [cv]
                    cf = [
                        [cv[2], cv[3], cv[7], cv[6]],  # back
                        [cv[0], cv[1], cv[5], cv[4]],  # front
                        [cv[0], cv[1], cv[2], cv[3]],  # Bottom
                        [cv[4], cv[5], cv[6], cv[7]],  # top
                        [cv[4], cv[7], cv[3], cv[0]],  # left
                        [cv[1], cv[2], cv[6], cv[5]],  # right
                    ]
                    cube_faces += [cf]
                    if self.cubes[x][y][z] == 0.0:
                        colors += ["cyan"]
                    else:
                        colors += ["r"]

        return (cube_verts, cube_faces, colors)

    def draw(self, mesh_verts, mesh_faces):
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        r = [-self.n, self.n]
        X, Y = np.meshgrid(r, r)
        (grid, faces, colors) = self.generate_grid()
        for f, c in zip(faces, colors):
            if c == "r":
                alpha = 0.25
                ax.add_collection3d(
                    Poly3DCollection(
                        f, facecolors=c, linewidth=1, edgecolors="r", alpha=alpha
                    )
                )
        for face in mesh_faces:
            f = (
                np.array(
                    [mesh_verts[face[0]], mesh_verts[face[1]], mesh_verts[face[2]]]
                ),
            )
            ax.add_collection3d(
                Poly3DCollection(
                    verts=f,
                    zsort="average",
                    facecolors="cyan",
                    linewidth=0.5,
                    edgecolors="r",
                    alpha=0.1,
                )
            )

        ax.set_xlim3d(0, n + 1)
        ax.set_ylim3d(0, n + 1)
        ax.set_zlim3d(0, n + 1)
        plt.xlabel("x")
        plt.ylabel("y")
        plt.show()

    def intersect_mesh(self, mesh):
        print("Generating intersections")
        verts = mesh.get_vertices()
        faces = mesh.get_faces()
        mutex = Lock()
        # Face based partitioning
        # We'll apply a face based partitioning scheme and add in logic for if (n_faces < max_threads), then we'll let faces generate more to fill it out.
        
        # First, we calculate our partition_size based on the n_max_threads & n_faces
        n_faces = len(faces)
        n_max_threads = self.max_t  # Just for visual clarity we'll rename it locally.
        part_size = math.floor(n_faces / n_max_threads)
        part_size = max(part_size, 1) # if n_max_threads > n_faces, part_size would be < 1.

        # Create partitions
        partitions = [faces[i: i + part_size] for i in range(0, n_faces, part_size)]
        # Overflow partition calc
        if(len(partitions) > n_max_threads):
          new_parts = partitions[:n_max_threads]
          # Stick overflow on the last index. We could spread it, but fuck it.
          if(len(partitions[n_max_threads:]) > 1):
            # Let's just nest this list comprehension and pretend it's not a double nested for loop
            overflow = np.array([p for p in [ z for z in partitions[n_max_threads:]]])
            overflow = overflow.reshape(overflow.shape[0]*overflow.shape[1],overflow.shape[2])
          else:
            overflow = partitions[-1]

          new_parts[-1] = np.append(new_parts[-1], overflow,axis=0)
          partitions = new_parts
        print("Created %d partitions of %d avg size" % (len(partitions), part_size))

        icount = Value('d', 0)
        # Create processes

        processes = []
        for partition in partitions:
          processes.append(Process(target=self.intersect_partition, args=(verts, partition, icount, mutex)))
          processes[-1].start()
        # Wait for all
        for p in processes:
          p.join()
        # connection.wait(p.sentinel for p in processes)
        return icount.value

    def intersect_partition(self, verts, part, count, mutex):
      """Threaded portion of the mesh intersection"""
      icount = 0
      for face in part:
        v = np.array([verts[face[0]], verts[face[1]], verts[face[2]]])
        icount += self.intersect_face(v)
      mutex.acquire()
      count.value += icount
      mutex.release()

    def intersect_face(self, verts):
        icount = 0
        icount = self.intersect_xy(verts)
        icount += self.intersect_xz(verts)
        icount += self.intersect_yz(verts)
        return icount

    def get_grid_scale(self):
        return float(self.n)

    # def center_mesh(self, mesh):
    #     verts = mesh.get_vertices()
    #     xmax = np.max(verts[:, 0])
    #     ymax = np.max(verts[:, 1])
    #     zmax = np.max(verts[:, 2])
    #     xmin = np.min(verts[:, 0])
    #     ymin = np.min(verts[:, 1])
    #     zmin = np.min(verts[:, 2])
    #     xdiff = abs(xmax - xmin)
    #     ydiff = abs(ymax - ymin)
    #     zdiff = abs(zmax - zmin)
    #     scale = max(xdiff, ydiff, zdiff)
    #     scale = self.n / scale
    #     scale = np.array([scale, scale, scale])
    #     mesh.set_scale(scale)
    #     center = (self.n / 2.0, self.n / 2.0, self.n / 2.0)
    #     mesh.set_position(center)



def check_blank_line(obj):
    """Determines if the inputs are blank"""
    valid = 1
    non_blank_inputs = [obj.axiom_edit, obj.angle_edit, obj.iters_edit]
    if obj.made_angle == True:
        non_blank_inputs.append(obj.turn_angle_edit)
    if obj.made_line == True:
        non_blank_inputs.append(obj.line_scale_edit)
    for rule in obj.prod_rules_edit:
        non_blank_inputs.append(rule)

    for input_box in non_blank_inputs:
        if input_box.valid == True and len(input_box.text()) == 0:
            print_error_message(input_box, "Blank input")
            valid = 0
    return valid


def check_in_alphabet(obj):
    """determines if the characters are in the alphabet"""
    valid = 1
    obj.alphabetic_inputs = [obj.axiom_edit]
    prod_input = []
    for rule in obj.prod_rules_edit:
        prod_input.append(rule)
    for input_box in prod_input:
        if input_box.valid == True:
            for ch in input_box.text():
                if obj.graphix == obj.three_d and (ch =='H' or ch =='h'):
                    print_error_message(input_box, ch+" not allowed for 3D")
                    valid =0
                if not ch in obj.alphabet and not ch in obj.ctrl_char and not ch == ':':
                    print_error_message(input_box, ch + " not in obj.alphabet")
                    valid = 0
    for input_box in obj.alphabetic_inputs:
        if input_box.valid == True:
            for ch in input_box.text():
                if obj.graphix == obj.three_d and (ch =='H' or ch =='h'):
                    print_error_message(input_box, ch+" not allowed for 3D")
                    valid =0
                if not ch in obj.alphabet and not ch in obj.ctrl_char:
                    print_error_message(input_box, ch + " not in alphabet")
                    valid = 0
    return valid


def check_if_numeric(obj):
    """determines if the inputs are numeric"""
    valid = 1
    numeric_inputs = [obj.angle_edit, obj.iters_edit]
    if obj.made_angle == True:
        numeric_inputs.append(obj.turn_angle_edit)
    if obj.made_line == True:
        numeric_inputs.append(obj.line_scale_edit)
    for input in numeric_inputs:
        if input.valid == True:
            try:
                float(input.text())
            except:
                print_error_message(input, "Not a number")
                valid = 0
    return valid


def check_valid_numeric(obj):
    """determines if the numbers are valid"""
    valid = 1
    angle_input = [obj.angle_edit]
    if obj.made_angle == True:
        angle_input.append(obj.turn_angle_edit)
    for input in angle_input:
        if input.valid == True and (float(input.text()) > 360 or float(input.text()) < -360):
            print_error_message(input, "Not a valid angle")
            valid = 0
    positive_input = [obj.iters_edit]
    if obj.made_line == True:
        positive_input.append(obj.line_scale_edit)
    for input in positive_input:
        if input.valid == True and float(input.text()) <= 0:
            print_error_message(obj.iters_edit, "Not a valid number of iterations")
            valid = 0
    return valid


def check_prod_rule_format(obj):
    """determines if the production rules are in the proper format"""
    valid = 1
    for input_box in obj.prod_rules_edit:
        if input_box.valid and not ':' in input_box.text():
            print_error_message(input_box, " Missing : ")
            valid = 0
    return valid


def check_branching(obj):
    """determines if there is branching and if it is in a valid syntax"""
    valid = 1
    stack = []
    if obj.axiom_edit.valid:
        for ch in obj.axiom_edit.text():
            if ch == '[':
                stack.append("[")
            if ch == ']':
                if len(stack) == 0:
                    print_error_message(obj.axiom_edit, "Must have matching [ ]")
                    valid = 0
                else:
                    stack.pop()
            if len(stack) > 0:
                print_error_message(obj.axiom_edit, "Must have matching [ ]")
                valid = 0

    for input_box in obj.prod_rules_edit:
        if input_box.valid:
            text = input_box.text().split(":")
            if '[' in text[0] or ']' in text[0]:
                print_error_message(input_box, "Can't have [ ] in key")
            stack = []
            for ch in text[1]:
                if ch == '[':
                    stack.append("[")
                if ch == ']':
                    if len(stack) == 0:
                        print_error_message(input_box, "Must have matching [ ]")
                        valid = 0
                    else:
                        stack.pop()
            if len(stack) > 0:
                print_error_message(input_box, "Must have matching [ ]")
                valid = 0
    return valid


def input_check(obj):
    """collects all the input checks and determines if they are valid"""
    valid = []

    valid.append(check_blank_line(obj))
    valid.append(check_in_alphabet(obj))
    valid.append(check_if_numeric(obj))
    valid.append(check_valid_numeric(obj))
    valid.append(check_prod_rule_format(obj))
    valid.append(check_branching(obj))
    valid.append(check_nd(obj))
    # reset valid
    obj.axiom_edit.valid = 1
    obj.angle_edit.valid = 1
    if obj.made_angle == True:
        obj.turn_angle_edit.valid = 1
    if obj.made_line:
        obj.line_scale_edit.valid = 1
    obj.iters_edit.valid = 1
    for input_box in obj.prod_rules_edit:
        input_box.valid = 1
    return (sum(valid) == len(valid))


def print_error_message(obj, msg):
    obj.setStyleSheet("color: red;")
    print("[ Error ] ", msg)
    obj.valid = False


def check_nd(obj):
    """determine if the percentages are valid"""
    valid = 1
    productions = dict()
    for rule in obj.prod_rules_edit:
        temprule = rule.text().split(":")
        productions[temprule[0]] = 0
    for i,rule in enumerate(obj.prod_rules_edit):
        if rule.valid:
            temprule = rule.text().split(":")
            try:
                productions[temprule[0]] = productions[temprule[0]] + (float(obj.prod_percent[i].text()))
            except:
                print("[ ERROR ] ",obj.prod_percent[i].text()," is not a number.")
                valid =0
                return valid
    for prod,perc in productions.items():
        if perc != 1.0:
            print("[ ERROR ] Invalid percent of ",perc," entered for ",prod)
            valid = 0
    return valid



"""This file is the run.py that makes the application run"""
import sys
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (QCheckBox, QToolBar, QLabel, QMainWindow, QApplication, QPushButton, QDialog,
                             QActionGroup,QStatusBar, QMenuBar, QAction, QHBoxLayout, QWidget, QGridLayout)

from PyQt5.QtWidgets import QAction, QApplication, QMainWindow, QMenu
from PyQt5.QtCore import QSize, Qt
from lsystem.core.getting_started import GettingStarted
from lsystem.core.glossary import Glossary
from lsystem.core.my_ui import UIWidget
from lsystem.core.save_rules_window import SaveRules
from lsystem.core.settings import PopupSettings

from lsystem.core.lsystem_2d_widget import LSystem2DWidget
from lsystem.core.lsystem_3d_widget import LSystem3DWidget

class MyMainWindow(QMainWindow):
    """This class is the main UI Window that is the parent for the entire application"""

    def __init__(self, parent=None):
        """Defaults the application window to be 500X500"""
        super(MyMainWindow, self).__init__(parent=parent)
        self.left = 500
        self.top = 500
        self.width = 500
        self.height = 500
        self.ui_widget = UIWidget()
        self.popup_settings = PopupSettings(self.ui_widget.graphix)
        self.glossary = Glossary()
        self.save_rules = SaveRules(self.ui_widget)
        self.getting_started = GettingStarted()
        self.init_window()

        toolbar = QToolBar("Settings Toolbar")
        toolbar.setIconSize(QSize(16,16))
        self.addToolBar(toolbar)
        toolbar.setMovable(False)
        toolbar.setContextMenuPolicy(Qt.PreventContextMenu)

        self.button_action = QAction("2D", self)
        self.button_action.setStatusTip("Click to use 2D L-Systems!")
        self.button_action.triggered.connect(self.toggle_dim2D)

        self.button_action2 = QAction("3D", self)
        self.button_action2.setStatusTip("Click to use 3D L-Systems!")
        self.button_action2.triggered.connect(self.toggle_dim3D)

        self.save_action = QAction('Save Rules', self)
        self.save_action.setStatusTip("Click to save the grammar of the L-System!")
        self.save_action.triggered.connect(lambda: self.save_rules.show())

        self.glossary_action = QAction("Glossary", self)
        self.glossary_action.setStatusTip("Click to see the glossary!")
        self.glossary_action.triggered.connect(lambda: self.glossary.show())

        self.tutorial = QAction("Tutorial", self)
        self.tutorial.setStatusTip("Click here to see how to get started!")
        self.tutorial.triggered.connect(lambda: self.getting_started.show())

        self.reset_zoom_button = QAction("Reset Zoom",self)
        self.reset_zoom_button.setStatusTip("Click here to reset zoom!")
        self.reset_zoom_button.triggered.connect(self.reset_zoom)

        self.screenshot_button = QAction("Screenshot",self)
        self.screenshot_button.setStatusTip("Click here to take a screenshot!")
        self.screenshot_button.triggered.connect(self.screenshot)

        toolbar.addAction(self.tutorial)
        toolbar.addAction(self.glossary_action)
        toolbar.addAction(self.save_action)
        toolbar.addAction(self.reset_zoom_button)
        toolbar.addAction(self.screenshot_button)

        toolbar.addSeparator()


        toolbar.addAction(self.button_action)
        toolbar.addAction(self.button_action2)

        self.button_action.setCheckable(True)
        self.button_action2.setCheckable(True)
        self.button_action.setChecked(True)

        self.setStatusBar(QStatusBar(self))

        self.setCentralWidget(self.ui_widget)


    def toggle_dim2D(self, s):
        if(s):
          print("I want 2D!")
          self.button_action2.setChecked(False)
          self.ui_widget.dims.setCurrentWidget(self.ui_widget.two_d)
          self.ui_widget.graphix = self.ui_widget.two_d
        else:
          self.button_action.setChecked(True)
          print("Nothing should happen :)")
    def toggle_dim3D(self, s):
        if(s):
            print("I want 3D!")
            self.button_action.setChecked(False)
            self.ui_widget.dims.setCurrentWidget(self.ui_widget.three_d)
            self.ui_widget.graphix = self.ui_widget.three_d
        else:
          self.button_action2.setChecked(True)
          print("Nothing should happen")


    def reset_zoom(self):
        self.ui_widget.reset_zoom()
    def screenshot(self):
        pos = self.pos()
        #pos.setY(self.pos().y()+self.height)
        self.ui_widget.screenshot(pos)

    def init_window(self):
        """Shows the main window"""
        self.setGeometry(self.left, self.top, self.width, self.height)
        self.setWindowTitle("L-System Generator")
        self.show()

    def close_event(self):
        """Makes sure everything gets properly deleated upon closing"""
        print("[ INFO ] Exiting...")
        sys.exit()


if __name__ == "__main__":
    APP = QApplication(sys.argv)
    DISPLAY = MyMainWindow()
    R = APP.exec_()
    sys.exit(R)



"""This file handles the OpenGL Window actions"""
from time import time

# Other includes
import numpy as np
# Lsystem includes
from glm import vec3
# Python core includes
from PIL import Image
from pyqtgraph import GraphItem, PlotWidget
from pyqtgraph.opengl import GLLinePlotItem, GLMeshItem, GLViewWidget
# PyQt5 includes
from PyQt5.QtWidgets import QApplication, QLabel, QVBoxLayout, QWidget

# from graphics.graph_mesh import GraphObject
from lsystem.core.graph import Graph

import pyscreenshot as ImageGrab

class LSystem2DWidget(PlotWidget):
    """
    This class manages the 2D PlotWidet used to display the lsystem.

    Attributes:
    mesh (MeshObject): The mesh object that is being displayed.
    graph (Graph): Digraph class used to minimize vector redundancy.
    """
    def __init__(self):
        """
        Constructor of the LSystem2DWidget
        """
        super(LSystem2DWidget, self).__init__()

        # Production scene objects.
        self.graph = Graph()
        self.mesh = None

    def reset_camera(self):
        """
        Resets the camera to the initial position & zoom level.
        """
        self.reset_zoom() # This happens to do the same thing.

    def reset_zoom(self):
        """
        Resets the camera to the initial position & zoom level.
        """
        self.autoRange()

    def screenshot(self, filename, pos):
        """
        Captures the view of the widget and saves it to a filename on disk.

        Parameters:
        filename (str): Filename to save to.
        pos (???): Nani the fuck is this shit?
        """
        #print("[ INFO ] Intentionally broken at the moment. There is a bug in pyqtgraph's image exporter so we'll need to fork or do it ourselves")
        #rect is the rectange (x,y,width,height) of the widget
        rect = self.frameGeometry()
        #shift over to get absl. pos
        rect.translate(pos)
        im = ImageGrab.grab(bbox=(rect.x(), rect.y(), rect.x()+rect.width(), rect.y()+rect.height())) # X1,Y1,X2,Y2
        im.save(filename)

    def set_graph(self, graph):
        """
        Sets the graph object the scene uses to generate the mesh.

        Parameters:
        graph (Graph): Digraph used to generate the mesh.
        """
        self.clear()
        self.graph = graph
        (verts, adj) = self.graph.export_to_pyqtgraph()
        self.mesh = GraphItem()
        self.mesh.setData(pos=verts, adj=adj, symbol=None)
        self.addItem(self.mesh)

    def clear(self):
        """
        Removes the mesh from the scene.
        """
        if isinstance(self.mesh, GraphItem):
            self.removeItem(self.mesh)
        self.mesh = None



from glm import (
    mat4,
    mat3,
    vec3,
    vec4,
    mat4_cast,
    normalize,
    conjugate,
    angleAxis,
    scale,
    translate,
)
import numpy as np
from scipy.spatial.transform import Rotation as R


class SpatialObject:
    def __init__(self):
        self.model_matrix = mat4(1.0)
        self.position = vec3(0.0)
        self.current_scale = vec3(1.0)
        self.rotation = R.from_rotvec(np.array([0., 0., 0.]))

    
    def get_rotation(self):
        return self.rotation

    def rotate(self, rotation):
        """applies a rotation using radians"""
        r = R.from_rotvec(np.array(rotation))
        self.rotation = r * self.rotation

    def set_rotation(self, rotation):
        """Sets our rotation in radians."""
        self.rotation = R.from_rotvec(np.array(rotation))

    def set_position(self, pos):
        self.position = vec3(pos)

    def get_position(self):
        return self.position

    def translate(self, offset, relative=True):
        offset = vec3(offset)
        if relative:
            self.position += self.get_x_axis() * offset.x
            self.position += self.get_y_axis() * offset.y
            self.position += self.get_z_axis() * offset.z
        else:
            self.position += offset

    # Generates and returns the object's current model matrix.
    def generate_model_matrix(self):
        trans_matrix = translate(mat4(1.0), self.position)
        scale_matrix = scale(mat4(1.0), self.current_scale)
        self.model_matrix = trans_matrix * scale_matrix
        return self.model_matrix

    def scale(self, s):
        self.current_scale *= s

    def set_scale(self, s):
        self.current_scale = vec3(s)



import sys
import math
from time import time
from copy import deepcopy

import numpy as np
import pywavefront as pwf

from lsystem.boxcounting3d.cubes import Cubes

# This entire setup assumes that we're working with the MeshObject used in the L'System Visualizer project

def calc_fractal_dim3D(mesh):
  """Wrapper function that calls calc_fractal_dim and narrows in on a good resolution for n"""

  # TODO testing
  # So in theory our accuracy should decrease if the cubes become smaller than the diameter of the smallest point of the mesh..
  # This would cause mesh parts to go unregistered. To prevent this, we can limit the cube side length to
  # 1 / n >= diameter. But this doesn't allow us to zoom in very far. For the pipe our diameter is 0.2...
  # This would seem to limit us to an n of 5 accurately?
  # This would limit us to 5 if we were doing a single instance of the mesh which would occupy a 1 unit length space.
  # But if we scale up, this value changes. If we have 2 pipes == side by side like this, then they're effectively scaled
  # to 1/2 size to fit. So our subdivision max becomes 1/2 * diamater? so our new max becomes 0.1.
  # We can ensure this factor to hold because we rescale the mesh by a factor of the greatest difference between
  # two points on the same axis. So a singular mesh, this factor would be 1x, two meshes laid length wise
  # would be 2 units long, or 2 units between the max two.
  # The question here is, is the factor we care about related to just the scale factor of n, or is it a
  # factor of the mesh itself? i.e. a dumb 1/n, or is it a ratio of how the mesh will be rescaled.
  # If it's a factor of the mesh x,y,z ratio, part of it would include the max delta of x y and z.
  radius = 0.1 # Hard coded for our pipe mesh. We'll need to adjust this if we want to support others.
  deltas = get_deltas(mesh)
  delta = max(deltas) # This is used both for our scale factor & for our min sidelength calc.
  # Min sidelength should be a function of 1/delta * diameter.
  # 0.1
  min_side = 1 / delta * 2 * radius
  max_n = int(1 / min_side)

# [(1, -1), (2, 3.0), (3, -1), (4, 2.0), (5, 1.7958889470453636), (6, 1.7737056144690833), (7, 1.8270874753469162),
# (8, 1.6666666666666667), (9, 1.9886214170079384), (10, 1.9344984512435675), (11, 1.8991141703557868), (12, 2.0),
# (13, 1.843072575667113), (14, 1.9415887339401865), (15, 1.8809659432212678), (16, 2.0), (17, 1.8972374314701483),
# (18, 1.959249866272526), (19, 2.020976693588105), (20, 2.004969940947389), (21, 1.98083127440203),
# (22, 2.02814952505589), (23, 2.0065637842976924), (24, 2.000545807561736), (25, 1.9878352824614485),
# (26, 2.0577703197939745), (27, 1.9928407195028168), (28, 2.023549459585051), (29, 2.043601646865157)]
  print("Min sidelength %f" % min_side)
  print("Max detected n: %f" % max_n)
  dims = []
  with open("somefile.db","w+") as f:
    for n in range(2, min(max_n+1,30)):
      dmesh = deepcopy(mesh)
      line = "n: {:d} s: {:f} | dim: {:f}".format(n, 1.0/n, calc_fractal_dim(n, dmesh, deltas))
      f.write(line)

 
  return 0

def get_deltas(mesh):
  """Calculates the deltas of x,y,z and returns a tuple of these values.
  """
  verts = mesh.get_vertices()
  xmax = np.max(verts[:, 0])
  ymax = np.max(verts[:, 1])
  zmax = np.max(verts[:, 2])
  xmin = np.min(verts[:, 0])
  ymin = np.min(verts[:, 1])
  zmin = np.min(verts[:, 2])
  delta_x = abs(xmax - xmin)
  delta_y = abs(ymax - ymin)
  delta_z = abs(zmax - zmin)
  return (delta_x, delta_y, delta_z)

def calc_fractal_dim(n: int, mesh, deltas):
  """Calculates the fractal dimensions for nxnxn cubes around the given mesh"""
  scale = center_mesh(mesh, n, deltas) # Center the mesh in teh cubic volume
  side_length = 1.0 / n  # We normalize the mesh, so the side length is important here.
  cubes = Cubes(n)
  start = time()
  intersections = cubes.intersect_mesh(mesh)
  end = time() - start
  print("Vertice count: %d" % len(mesh.get_vertices()))
  print("Faces in the mesh: %d" % len(mesh.get_faces()))
  print("N-Cubes: %d" % n**3)
  print("Side length: %f" % side_length)
  if(intersections > 0):
    print("Intersection count %d" % intersections)
    dim = math.log(intersections) / math.log(1 / side_length)
    print("Fractal dimension: " + str(dim))
  else:
    print("No intersections detected")
    dim = -1
  print("Took %f seconds" % end)
  return dim

def center_mesh(mesh, n, deltas):
  """Centers the 3d mesh in the space of our cube grid"""
  scale = max(deltas)
  scale = n / scale
  scale = np.array([scale, scale, scale])
  mesh.set_scale(scale)
  center = (n / 2.0, n / 2.0, n / 2.0)
  mesh.set_position(center)
  return scale

# def main():
#   # model = 'SierpinskiTetrahedron.obj'
#   model = 'MengerSponge.obj'
#   scene = pwf.Wavefront(model, collect_faces=True)

#   faces = np.array(scene.mesh_list[0].faces)
#   vertices = np.array(scene.vertices)

#   mesh = MeshObject(vertexes=vertices, indices=faces)
#   # calc_fractal_dim(10, mesh)
#   # calc_fractal_dim(20, mesh)
#   calc_fractal_dim(30, mesh)
#   # calc_fractal_dim(40, mesh)
#   # calc_fractal_dim(50, mesh)
#   # calc_fractal_dim(100, mesh)
#   # calc_fractal_dim(10, mesh)
#   # calc_fractal_dim(20, mesh)




# This file is for our undirected graph class and helper functions.
# A few notes on the different design decisions here.
# 1. Using a python dictionary with vertices as a key lets us check if vertices exist in much quicker time than doing some_list.index(value).
# 2. Using tuples for vertices allows them to be hashable, aka used as keys in a dict.
# 3. By storing each edge between v1&v2 only once in v1's adjacency_list, we avoid any possible concurrency issues.
# 4. I'm pretty sure we avoid global interpreter lock with this implementation of threading. Will need thorough testing.
from time import time

import numpy as np

from lsystem.graphics.colors import Colors


class Graph:
    """
    This class is an undirected graph implementation with utility methods for exporting to different mesh formats.

    Attributes:
    adjacency_list (dict): Dictionary of vertices and their connected partners.
    colors (list): List of vertex colors.
    vertices (list): List of vertices.
    """
    def __init__(self):
        """
        Constructor for the Graph class.
        """
        # Vertices will be a dictionary that stores an adjacency list for each vertex.
        self.adjacency_list = {}
        self.colors = []
        self.vertices = []
        self.n = 0

    # Let's assume the vertex passed is a tuple since that's hashable and we need that for dictionaries to work.
    # Optional color passed for each vertex. Default is white.
    def add_vertex(self, vertex, color=Colors.white):
        """
        Adds a vertex to the undirected graph.

        Parameters:
        vertex (vec3): Vertex to add.
        color (vec3): Color of the vertex.
        """
        if vertex not in self.adjacency_list:
            self.adjacency_list[vertex] = {"edges": [], "color": color, "index": self.n}
            self.vertices.append(vertex)
            self.colors.append(color)
            self.n += 1

    # Assume both v1 and v2 are tuples of form (x,y,z)
    # We'll just add to v1 each time.
    def add_edge(self, v1, v2):
        """
        Adds an edge between two vertices. Will add the vertices if they don't already exist in the graph.

        Parameters:
        v1 (vec3): Vertex position 1
        v2 (vec3): Vertex position 2
        """
        if v1 not in self.adjacency_list:
            self.add_vertex(v1)
        if v2 not in self.adjacency_list:
            self.add_vertex(v2)
        if not self.edge_exists(v1, v2):
            self.adjacency_list[v1]["edges"].append(v2)

    def clear(self):
        """
        Clears the graph.
        """
        self.adjacency_list = {}
        self.colors = []
        self.vertices = []

    # Checks if an edge exists.
    def edge_exists(self, v1, v2):
        """
        Checks if an edge exists between two vertices.

        Parameters:
        v1 (vec3): Vertex 1
        v2 (vec3): Vertex 2

        Returns:
        - True/False for if it exists.
        """
        if (
            v1 in self.adjacency_list[v2]["edges"]
            or v2 in self.adjacency_list[v1]["edges"]
        ):
            return True
        return False

    def export_to_ogl_verts(self):
        """
        Exports the graph vertices to a format OpenGL can use.

        Returns:
        - numpy.array: A numpy array of the vertices.
        """
        verts = []
        for v in self.vertices:
            for e in self.adjacency_list[v]["edges"]:
                verts += [v, e]
        return np.array(verts)

    def export_to_matplot2d(self):
        """
        Exports the graph to vertices matplotlib can use.

        Returns:
        - list: List of vertices.
        """
        lines = []
        for v in self.vertices:
          for e in self.adjacency_list[v]["edges"]:
            lines += [((v[0], e[0]), (v[1], e[1]))]
        return lines

    def export_to_pyqtgraph(self):
        """
        Exports the graph to a format MeshObject can use for pyqtgraph's 3D mesh objects.

        Returns:
        - tuple: (verts, adj)
        """
        # Reshape in case we have 3 dimensional data
        verts = np.array(self.vertices)
        verts = verts[:, :2]
        adj = []
        for v in self.vertices:
          for e in self.adjacency_list[v]["edges"]:
            adj += [(self.adjacency_list[v]["index"], self.adjacency_list[e]["index"])]
        adj = np.array(adj)
        return (verts, adj)

    def generate_indices(self, adjacency_list, vert_list):
        """
        Generates the indices of the vertices/adjacency list for OpenGL element drawing calls.

        Parameters:
        adjacency_list (dict): Dictionary of vertex->vertex mappings
        vert_list (list): List of all vertices.

        Returns:
        - list: List of indices.
        """
        # for each vertice in our vertice range
        #   for every outgoing edge
        #       add that line's indice.
        start = time()
        inds = []
        for v in vert_list:
            for e in adjacency_list[v]["edges"]:
                inds += [adjacency_list[v]["index"], adjacency_list[e]["index"]]
        end = time()
        print(
            "[ INFO ] Finished exporting to OGL indices in "
            + str(round(end - start, 3))
            + "s"
        )
        return inds

    # this function will generate opengl vertex data from a graph.
    def export_to_ogl_indices(self):
        """
        Generates & exports all vertex, color, and indice data needed for OpenGL Element drawing calls.

        Returns:
        - tuple(np.array, np.array, np.array)
        - - Array 0 is the array of vertices.
        - - Array 1 is the array of colors.
        - - Array 2 is the array of indices.
        """
        adjacency_list = self.adjacency_list
        vertices = np.array(self.vertices)
        colors = self.colors
        indices = self.generate_indices(adjacency_list, self.vertices)
        vertices = vertices.reshape(vertices.shape[0] * vertices.shape[1])
        return (np.array(vertices), np.array([0, 0, 0, 1]), np.array(indices))



"""
This file adds the graphics to the UI
F(and G) draws a unit length line
f(and g) moves forward a unit length
H draws a half length line
h moves forward half a unit length
- turns counter-clockwise
+ turns clockwise
[ starts branch
] ends branch
| reverses direction
( decrements the angle by a turning angle
) increments the angle by a turning angle
"""

import matplotlib.pyplot as plt
import numpy as np
from PyQt5.QtWidgets import QStackedWidget, QVBoxLayout, QPushButton, QWidget, QGridLayout, QLabel, QFileDialog
from PyQt5 import QtWidgets, QtCore
from PyQt5.QtCore import Qt
from lsystem.core.lsystem_utils import (
    load_saved_lsystems,
    get_saved_lsystem,
    generate_lsystem_2d,
    generate_lsystem_3d,
)
from PyQt5.QtGui import QIcon
from lsystem.core.input_check import input_check

from lsystem.core.lsystem_2d_widget import LSystem2DWidget
from lsystem.core.lsystem_3d_widget import LSystem3DWidget

#from lsystem.core.fractal_dim import fractal_dim_calc
from lsystem.core.fract_menu import FractalDimension, fractal_dim_calc
from lsystem.boxcounting3d.calc import calc_fractal_dim3D
import copy

import os

class CustomLineEdit(QtWidgets.QLineEdit):
    """ Class that enables input in a text box """

    clicked = QtCore.pyqtSignal()
    def __init__(self):
        super().__init__()
        self.valid = True
        self.error_message = "X"

    # Needs to be CamelCase
    def mousePressEvent(self, QMouseEvent):
        self.clicked.emit()

    def reset_color(self):
        self.setStyleSheet("color: black;")

    def clear_box(self):
        self.setText("")
        self.setStyleSheet("color: black;")

    def reset_box(self):
        self.reset_color()
        self.clear_box()



class UIWidget(QWidget):
    """ Class that holds all of the widgets for viewing """

    def __init__(self):
        """ Initializes class and variables """
        super(UIWidget, self).__init__()
        self.axiom = QLabel("Axiom")
        self.angle = QLabel("Angles(degrees)")
        self.iters = QLabel("Iterations")
        self.axiom_edit = CustomLineEdit()
        self.angle_edit = CustomLineEdit()
        self.iters_edit = CustomLineEdit()
        self.text_boxes = [self.axiom_edit, self.angle_edit, self.iters_edit]
        self.prod_plus = QPushButton("+", self)
        self.lsys_button = QPushButton("Generate L System", self)
        self.boxcount_button = QPushButton("Fractal Dim", self)
        
        self.widget = QWidget()
        self.scroll_area = QtWidgets.QScrollArea()
        self.layout_examples = QVBoxLayout(self.widget)

        self.prods = 1
        self.prod_rules_edit = []
        self.examples = []
        self.minuses = None
        self.made_angle = False
        self.prod_percent = []
        self.amount = 0
        self.index = 0
        self.frac_points = []
        self.made_line = False
        self.prod_rules = []
        self.verts = []  # This will store the vertices from generate_lsystem
        self.saved_lsystems = load_saved_lsystems()
        self.two_d = LSystem2DWidget()
        self.three_d = LSystem3DWidget()

        self.fractal_menu = FractalDimension(self)
        self.dims = QStackedWidget()
        self.dims.addWidget(self.two_d)
        self.dims.addWidget(self.three_d)
        self.dims.setCurrentWidget(self.two_d)
        self.graphix = self.two_d

        self.init_UI()
        self.alphabet = [
            "F",
            "f",
            "G",
            "g",
            "H",
            "h",
            "^",
            "&",
            "-",
            "+",
            "[",
            "]",
            "|",
            "(",
            ")",
            ">",
            "<",
            " ",

        ]
        self.ctrl_char = [
            "A",
            "B",
            "C",
            "D",
            "E",
            "I",
            "J",
            "K",
            "L",
            "M",
            "N",
            "O",
            "P",
            "Q",
            "R",
            "S",
            "T",
            "U",
            "V",
            "W",
            "X",
            "Y",
            "Z",
        ]

    def init_UI(self):
        """ Creates and adds all widgets in the viewport and sets the layout  """
        # renames the window
        self.setWindowTitle("L-Systems Generator")
        self.layout = QGridLayout()
        self.init_buttons()
        self.init_text_boxes()
        self.add_widgets()
        self.setLayout(self.layout)
        self.setGeometry(500, 500, 500, 500)

    def init_text_boxes(self):
        '''Links all of the text boxes to buttons'''
        # creates the labels for each text box
        self.prod_rules.append(QLabel("Production Rule " + str(self.prods)))

        # creates the text box for each label
        self.axiom_edit.returnPressed.connect(self.lsys_button.click)
        self.axiom_edit.clicked.connect(lambda: self.axiom_edit.reset_color())

        self.prod_rules_edit.append(CustomLineEdit())
        self.prod_rules_edit[0].clicked.connect(
            lambda: self.prod_rules_edit[0].reset_color()
        )
        self.text_boxes.append(self.prod_rules_edit[-1])
        self.prod_rules_edit[0].returnPressed.connect(self.lsys_button.click)
        self.prod_rules_edit[0].textChanged.connect(lambda: self.show_popup())

        self.prod_percent.append(CustomLineEdit())
        self.prod_percent[0].setFixedWidth(50)
        self.prod_percent[0].setText("1")

        self.text_boxes.append(self.prod_percent[-1])
        self.angle_edit.returnPressed.connect(self.lsys_button.click)
        self.angle_edit.clicked.connect(lambda: self.angle_edit.reset_color())

        self.iters_edit.returnPressed.connect(self.lsys_button.click)
        self.iters_edit.clicked.connect(lambda: self.iters_edit.reset_color())

        self.prod_plus.clicked.connect(self.more_prods)

    def init_buttons(self):

        # makes the lsys generator button
        self.lsys_button.clicked.connect(self.on_lsys_button_clicked)
        self.lsys_button.setAutoDefault(True)

        self.boxcount_button.clicked.connect(self.on_boxcount_button_clicked)
        self.boxcount_button.setAutoDefault(True)

        self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setFixedWidth(150)
        self.scroll_area.setWidget(self.widget)

        precons = ['SierpinksiTriangle', 'KochCurve', 'KochSnowflake',
            'KochIsland', 'PeanoCurve', 'DragonCurve', 'HilbertCurve',
            'TreeExample', 'IslandsandLakes']

        for i, key in enumerate(self.saved_lsystems):
            self.examples.append(QPushButton())
            if i < len(precons):
              self.examples[i].setIcon(QIcon('{}/lsystem/assets/images/{}.png'.format(os.getcwd(), precons[i])))
              self.examples[i].setIconSize(QtCore.QSize(120, 100))
            else:
              self.examples[i].setText(key)
            self.examples[i].clicked.connect(
                lambda state, x=key: self.gen_example(str(x))
            )
            self.layout_examples.addWidget(self.examples[i])
        self.layout_examples.addStretch(1)

    @QtCore.pyqtSlot()
    def on_lsys_button_clicked(self):
        self.gen_sys()

    def boxcount_2d(self):
        """2 dimensional fractal dimension calculation code."""
        self.fractal_menu.show()
        start_size = 8
        num_sizes = 7
        self.x_arr = []
        self.y_arr = []
        fract_avg = []
        end_size = start_size * (2 ** num_sizes)
        temp_verts = copy.deepcopy(self.verts)
        fractal_dim = fractal_dim_calc(temp_verts, end_size, num_sizes)
        for i in range(num_sizes):
            self.x_arr.append(np.log2((start_size)))
            self.y_arr.append(fractal_dim[i])

            fract_avg.append(np.polyfit(self.x_arr, self.y_arr, 1)[0])
            print("(box width = 1/", start_size, ") FRACTAL AVG: ", fract_avg[-1])
            start_size = start_size * 2
        # y_arr = np.asarray(y_arr)
        # x_arr = np.asarray(x_arr)
        print("Made it this far")
        figi, ax = plt.subplots()
        line, = ax.plot(self.x_arr, self.y_arr, "bo", picker=5)
        ax.set_title(
            "Fractal dimension = {}".format(np.polyfit(self.x_arr, self.y_arr, 1)[0])
        )  # np.average(fract_avg)))
        figi.canvas.mpl_connect('pick_event', self.onpick1)
        #figi.show()
        plt.show()
        print("AVERAGE: ", np.average(fract_avg))

    def boxcount_3d(self):
        """3 dimensional fractal dimension calc code. I'll have to do some data sanitization somewhere. Maybe here."""
        mesh = self.graphix.mesh
        if(mesh is not None):
            calc_fractal_dim3D(mesh)


    def is_2d(self):
        """Returns true if 2D is loaded, false if 3D is loaded."""
        if (self.dims.currentWidget().__class__.__name__ == 'LSystem3DWidget'):
            return False
        return True

    def on_boxcount_button_clicked(self):
        if(self.is_2d()):
            self.boxcount_2d()
        else:
            self.boxcount_3d()

    def onpick1(self, event):
        print("I am an EVENT1")
        #plt.close(fig=None)
        ind = event.ind[0]+1
        for i in range(ind):
            if(len(self.x_arr)==2):
                break
            # always delete index 0 because array shifts left after delete
            self.x_arr = np.delete(self.x_arr, 0, axis=None)
            self.y_arr = np.delete(self.y_arr, 0, axis=None)
        print("The x array is: ", self.x_arr)
        fig1, ax = plt.subplots()
        ax.plot(self.x_arr, self.y_arr, "bo", picker=5)
        ax.set_title(
            "Fractal dimension = {}".format(np.polyfit(self.x_arr, self.y_arr, 1)[0])
        )  # np.average(fract_avg)))
        fig1.canvas.mpl_connect('pick_event', self.onpick2)
        fig1.show()
        return True
    def onpick2(self, event):
        print("I am an EVENT2")
        #plt.close(fig=None)
        ind = event.ind[0]
        num_to_remove = len(self.x_arr)-ind
        for i in range(num_to_remove):
            if(len(self.x_arr)==2):
                break
            # always delete index -1
            self.x_arr = np.delete(self.x_arr, -1, axis=None)
            self.y_arr = np.delete(self.y_arr, -1, axis=None)
        fig2, ax = plt.subplots()
        ax.plot(self.x_arr, self.y_arr, "bo", picker=5)
        ax.set_title(
            "Fractal dimension = {}".format(np.polyfit(self.x_arr, self.y_arr, 1)[0])
        )  # np.average(fract_avg)))
        fig2.show()
        return True



    def add_widgets(self):
        # Adding widgets to window
        self.layout.addWidget(self.axiom, 1, 0)
        self.layout.addWidget(self.axiom_edit, 1, 1, 1, 10)
        self.layout.addWidget(self.prod_rules[0], 2, 0, 1, 1)
        self.layout.addWidget(self.prod_rules_edit[0], 2, 1, 1, 9)
        self.layout.addWidget(self.prod_percent[0], 2, 9)
        self.layout.addWidget(self.prod_plus, 2, 10, 1, 1)
        self.layout.addWidget(self.angle, 10, 0)
        self.layout.addWidget(self.angle_edit, 10, 1, 1, 10)
        self.layout.addWidget(self.iters, 13, 0)
        self.layout.addWidget(self.iters_edit, 13, 1, 1, 10)
        self.layout.addWidget(self.scroll_area, 14, 0, 1, 1)
        self.layout.addWidget(self.boxcount_button, 16, 0, 1, 1)
        self.layout.addWidget(self.dims, 14, 1, 5, -1)
        self.layout.addWidget(self.lsys_button, 20, 0, 1, -1)

    def show_popup(self):
        self.reset_text_box_color()
        prod_rule = ""
        rules = ""
        self.amount = 0
        for prod in self.prod_rules_edit:
            prod_rule += prod.text()
            temp = prod.text()
            temp = temp.replace(" ", "")
            temp = temp[:].split(":")[0]
            rules += temp
            rules += " "

        all_prod_rule = prod_rule + self.axiom_edit.text()

        if (")" in all_prod_rule or "(" in all_prod_rule) and self.made_angle is False:
            self.turn_angle = QLabel("Turning Angle")
            self.turn_angle_edit = CustomLineEdit()
            self.text_boxes.append(self.turn_angle_edit)
            self.turn_angle_edit.returnPressed.connect(self.lsys_button.click)
            self.turn_angle_edit.clicked.connect(
                lambda: self.turn_angle_edit.reset_color()
            )
            self.layout.addWidget(self.turn_angle, 11, 0)
            self.layout.addWidget(self.turn_angle_edit, 11, 1, 1, 10)
            self.made_angle = True

        if (
            self.made_angle is True
            and not "(" in all_prod_rule
            and not ")" in all_prod_rule
            and self.made_angle is True
        ):
            self.text_boxes.remove(self.turn_angle_edit)
            self.layout.removeWidget(self.turn_angle_edit)
            self.layout.removeWidget(self.turn_angle)
            self.turn_angle.deleteLater()
            self.turn_angle_edit.deleteLater()
            self.turn_angle_edit = None
            self.turn_angle = None
            self.made_angle = False

        if (">" in all_prod_rule or "<" in all_prod_rule) and self.made_line is False:
            self.line_scale = QLabel("Line Scale")
            self.line_scale_edit = CustomLineEdit()
            self.text_boxes.append(self.line_scale_edit)
            self.line_scale_edit.returnPressed.connect(self.lsys_button.click)
            self.line_scale_edit.clicked.connect(
                lambda: self.line_scale_edit.reset_color()
            )
            self.layout.addWidget(self.line_scale, 12, 0)
            self.layout.addWidget(self.line_scale_edit, 12, 1, 1, 10)
            self.made_line = True

        if (
            self.made_line is True
            and not "<" in all_prod_rule
            and not ">" in all_prod_rule
            and self.made_line is True
        ):
            self.text_boxes.remove(self.line_scale_edit)
            self.layout.removeWidget(self.line_scale_edit)
            self.layout.removeWidget(self.line_scale)
            self.line_scale.deleteLater()
            self.line_scale_edit.deleteLater()
            self.line_scale_edit = None
            self.line_scale = None
            self.made_line = False

    # Probably doesn't need self as a param, can just be static.
    # Generates a rule dictionary from an array of production rule strings taken from the UI
    def gen_rule_dict(self, prod_rules):
        """
        formats production rules as
        {"F": [[p,rule],[p,rule]], "f":[[p,rule],[p,rule]] ... }
        """
        rules = {}
        for rule in prod_rules:
            rule = rule.text()
            rule = rule.replace(" ", "")
            prod = rule.split(":")
            rules[prod[0]] = []
        for i, rule in enumerate(prod_rules):
            rule = rule.text()
            rule = rule.replace(" ", "")
            prod = rule.split(":")
            rules[prod[0]].append([float(self.prod_percent[i].text()), prod[1]])
        return rules

    def close_event(self):
        print("[ INFO ] Exiting...")
        exit()

    def more_prods(self):
        """ Creates more productions when + button is clicked """
        if self.prods < 4:
            self.prods = self.prods + 1
            self.prod_rules.append(QLabel("Production Rule " + str(self.prods)))
            self.prod_rules_edit.append(CustomLineEdit())
            self.prod_percent.append(CustomLineEdit())
            self.text_boxes.append(self.prod_rules_edit[-1])
            self.text_boxes.append(self.prod_percent[-1])
            self.prod_percent[-1].setFixedWidth(50)
            self.prod_rules_edit[self.prods - 1].textChanged.connect(
                lambda: self.show_popup()
            )
            self.prod_rules_edit[-1].returnPressed.connect(self.lsys_button.click)
            self.prod_rules_edit[-1].clicked.connect(
                lambda: self.prod_rules_edit[-1].reset_color()
            )
            self.layout.addWidget(self.prod_rules[self.prods - 1], self.prods + 1, 0)
            self.layout.addWidget(
                self.prod_rules_edit[self.prods - 1], self.prods + 1, 1, 1, 9
            )
            self.layout.addWidget(self.prod_percent[self.prods - 1], self.prods + 1, 9)

            if self.minuses is not None:
                # remove last minueses
                self.layout.removeWidget(self.minuses)
                self.minuses.deleteLater()
                self.minuses = None

            self.minuses = QPushButton("-", self)
            self.minuses.clicked.connect(self.less_prods)
            self.layout.addWidget(self.minuses, self.prods + 1, 10, 1, 1)
            self.prod_percent[-1].setText("1")

    def less_prods(self):
        """ Removes productions when - button is clicked """
        if self.prods > 1:
            self.text_boxes.remove(self.prod_rules_edit[-1])
            self.text_boxes.remove(self.prod_percent[-1])

            # remove last widget prodrules
            self.layout.removeWidget(self.prod_rules[-1])
            self.prod_rules[-1].deleteLater()
            self.prod_rules.pop()
            # remove last widget prodrulesEdit
            self.layout.removeWidget(self.prod_rules_edit[-1])
            self.prod_rules_edit[-1].deleteLater()
            self.prod_rules_edit.pop()
            # remove last percentage
            self.layout.removeWidget(self.prod_percent[-1])
            self.prod_percent[-1].deleteLater()
            self.prod_percent.pop()
            # remove last percentage
            # for i in self.index:
            #    for j in i:
            #        if j == self.prods-1:
            #             print("WE NEED TO DELETE")
            #             print(i)
            #      print("HELLO")
            #      self.layout.removeWidget(self.prodPercent[-1])
            #      self.prodPercent[-1].deleteLater()
            #      self.prodPercent.pop()
            #      self.amount = self.amount - 1
            #      print(len(self.prodPercent))
            # remove last minuses
            self.layout.removeWidget(self.minuses)
            self.minuses.deleteLater()
            self.minuses = None
            self.prods = self.prods - 1


        if self.prods > 1:
            self.minuses = QPushButton("-", self)
            self.minuses.clicked.connect(self.less_prods)
            self.layout.addWidget(self.minuses, self.prods + 1, 10, 1, 1)

    def gen_sys(self):

        """
        If the input is valid, iterates through productions
        and sends to graphics to be drawn
        """
        if input_check(self):
            axiom_input = self.axiom_edit.text()
            # prodInput = [self.prodrlesEdit.text()] #changed to array
            angle_input = self.angle_edit.text()
            if self.made_angle:
                turn_angle_input = self.turn_angle_edit.text()
            else:
                turn_angle_input = 0
            if self.made_line:
                line_scale_input = self.line_scale_edit.text()
            else:
                line_scale_input = 1
            iters_input = self.iters_edit.text()
            # Format input for use
            rules = self.gen_rule_dict(self.prod_rules_edit)
            # Generate rule grammar dictionary.
            grammar = {
                "rules": rules,
                "axiom": axiom_input,
                "iterations": int(iters_input),
                "angle": float(angle_input),
                "turnAngle": float(turn_angle_input),
                "lineScale": float(line_scale_input),
            }
            if (self.dims.currentWidget().__class__.__name__ == 'LSystem3DWidget'):
              self.mesh = generate_lsystem_3d(grammar)
              self.graphix.clear()
              self.graphix.add_mesh(self.mesh)
            else:
              self.verts = generate_lsystem_2d(grammar)
              # Sets verts on graphics widget and draws
              self.graphix.clear()
              self.graphix.set_graph(self.verts)
            # for i in range(1,len(self.verts)):
            #  self.graphix.set_graph(self.verts[i],1) #split = true
        self.graphix.update()
        self.graphix.reset_camera()

    def gen_example(self, example):
        self.axiom_edit.reset_box()
        for prod in self.prod_rules_edit:
            prod.reset_box()
        self.angle_edit.reset_box()
        self.iters_edit.reset_box()
        grammar = get_saved_lsystem(example, self.saved_lsystems)
        self.axiom_edit.setText(grammar["axiom"])

        num_rules = 0
        for key in grammar["rules"]:
            if isinstance(grammar["rules"][key], list):
                num_rules += len(grammar["rules"][key])
            else:
                num_rules += 1

        while self.prods < num_rules:
            self.more_prods()

        while self.prods > num_rules:
            self.less_prods()

        which_rule = 0
        for i, key in enumerate(grammar["rules"]):
            value = grammar["rules"][key]
            if isinstance(value, str):
                self.prod_rules_edit[which_rule].setText(key + ": " + value)
                which_rule+=1
            else:
                for val in value:
                    print(val)
                    self.prod_rules_edit[which_rule].setText(key + ": " + val[0])
                    self.prod_percent[which_rule].setText(val[1])
                    which_rule+=1

        self.angle_edit.setText(str(grammar["angle"]))
        if self.made_angle:
            self.turn_angle_edit.setText(str(grammar["turn_angle"]))
        if self.made_line:
            self.line_scale_edit.setText(str(grammar["line_scale"]))
        self.iters_edit.setText(str(grammar["iterations"]))
        self.gen_sys()

    def reset_text_box_color(self):
        for box in self.text_boxes:
            box.reset_color()
    def reset_zoom(self):
        self.two_d.reset_zoom()
        self.three_d.reset_zoom() #built in function don't change to snake script
    def screenshot(self, parent_pos):
        #rel pos is the upper left pointof the widget relative to window
        rel_pos = self.dims.pos()
        #shift the y down by the total height - height of the widget (height of text boxes)
        #1.1 factor added arbitrarily
        rel_pos.setY((self.height()-self.dims.height())*1.1)
        pos =  rel_pos+parent_pos
        qfd = QFileDialog()
        filter = "Images (*.png *.xpm *.jpg)"
        filename, type = QFileDialog.getSaveFileName(self, "", "", filter)
        if filename:
            self.two_d.screenshot(filename,pos)




'''This file performs all of the parsing'''
import random
import copy
import multiprocessing
import textwrap


def nd_choice(prods):
    '''Randomly selects the rule in nondeterminism'''
    prods_temp = copy.deepcopy(prods)
    rand = random.random()  # generate a random number in [0,1]
    for i, prod in enumerate(prods_temp):
        if rand > prod[0]:
            prods_temp[i + 1][0] += prod[0]
        else:
            return prod[1]

def string_parse(strings, prod_rules):
    """
    NOTE: Context sensitive reads left to right, with longest key getting priority

    str_temp is the current string
    it will be used to keep track of what parts of the string still need proccessing
    Context free implimentation idea:
    1. sort the keys from longest to shortest
    2. say the key is length n, if the first n letters of the string match the key, 
    add prod[key] to the new string
    3a.if the first n letters ever match a key remove the first n letters of the string
    3b.default rule: a letter goes to itself
    """
    newstr = ""
    str_temp = copy.deepcopy(strings)
    while len(str_temp) > 0:
        found = 0
        for key in sorted(prod_rules, key=len, reverse=True):
            if key == str_temp[: len(key)]:
                newstr += nd_choice(prod_rules[key])
                str_temp = str_temp[len(key) :]
                found = 1
                break
        # default prod is a letter goes to itself
        if found == 0:
            newstr += str_temp[0]
            str_temp = str_temp[1:]
    return newstr


def parsed_thread(strings, prods, iters):
    """
    Takes in an axiom set of prods and number of iterations and generates the new string
    """
    if max(list(prods.keys())) == 1:
        context_free = 1
    else:
        context_free = 0
    count = 0
    while count < iters:

        if (len(strings) > 1) and context_free:
            strs = textwrap.wrap(strings, int(len(strings) / cpus) + 1)
            strings = ""
            tmp = []
            cpus = multiprocessing.cpu_count()
            for i in range(cpus - 1):
                tmp[i] = multiprocessing.Process(
                    target=lambda x, arg1: x.put(
                        string_parse(arg1), args=(newstr1, strs[i], prods)
                    )
                )
                tmp[i].start()
            for i in range(cpus - 1):
                tmp[i].join()
                strings += newstr1.get()
        else:
            strings = string_parse(strings, prods)
        count += 1
    return strings


if __name__ == "__main__":
    TEST = [[0.2, "X"], [0.2, "Y"], [0.6, "Z"]]
    print(nd_choice(TEST))



import numpy as np
import math
from glm import vec3
import glm

from lsystem.graphics.mesh import MeshObject
from scipy.spatial.transform import Rotation as R

class Pipe(MeshObject):
  def __init__(self, start: vec3 = None, end: vec3 = None, rotation=vec3(0.0)):
    pos = np.multiply(np.add(start,end),.5)
    up = vec3(0,1,0)
    if(vec3(end-start) != up and vec3(end-start) != up*-1):
      rotation = glm.quatLookAt(vec3(end-start), up)
      MeshObject.__init__(self, position=pos, rot_quat=rotation)
    else:
      up = vec3(0,0,1)
      rotation = glm.quatLookAt(vec3(end-start), up)
      MeshObject.__init__(self, position=pos, rot_quat=rotation)


     
    verts = np.array([(0, 0, -0.5), (0, 0, 0.5), (0.1, 0.0, 0.5), (0.07071067811865477, 0.07071067811865475, 0.5), (6.123233995736766e-18, 0.1, 0.5), (-0.07071067811865475, 0.07071067811865477, 0.5), (-0.1, 1.2246467991473533e-17, 0.5), (-0.07071067811865477, -0.07071067811865475, 0.5), (-1.8369701987210297e-17, -0.1, 0.5), (0.07071067811865474, -0.07071067811865477, 0.5), (0.1, 0.0, -0.5), (0.07071067811865477, 0.07071067811865475, -0.5), (6.123233995736766e-18, 0.1, -0.5), (-0.07071067811865475, 0.07071067811865477, -0.5), (-0.1, 1.2246467991473533e-17, -0.5), (-0.07071067811865477, -0.07071067811865475, -0.5), (-1.8369701987210297e-17, -0.1, -0.5), (0.07071067811865474, -0.07071067811865477, -0.5)]
)
    faces = np.array([(0, 9, 2), (1, 17, 10), (0, 2, 3), (1, 10, 11), (0, 3, 4), (1, 11, 12), (0, 4, 5), (1, 12, 13), (0, 5, 6), (1, 13, 14), (0, 6, 7), (1, 14, 15), (0, 7, 8), (1, 15, 16), (0, 8, 9), (1, 16, 17), (17, 2, 9), (17, 10, 2), (10, 3, 2), (10, 11, 3), (11, 4, 3), (11, 12, 4), (12, 5, 4), (12, 13, 5), (13, 6, 5), (13, 14, 6), (14, 7, 6), (14, 15, 7), (15, 8, 7), (15, 16, 8), (16, 9, 8), (16, 17, 9)]
)
    self.set_vertexes(np.array(verts), False)
    self.set_indices(np.array(faces))
    self.start = start
    self.end = end

  def __reduce__(self):
    """Reduction code allowing us to pickle this class, thus allowing it to be deepcopied"""
    return (Pipe, (self.start, self.end))


  def _gen_pipe(self):
    """Function I used to create the verts for our pipe. I dumped these to a file to hardcode them to save runtime costs."""
    self.radius = 0.1
    self.divs = 8
    # self.height = 3 * self.radius
    self.height = 1

    # Generate circles for top and bottom of the pipe
    bcircle = []
    tcircle = []
    cstep = 2 * np.pi / self.divs
    for d in range(self.divs):
      v = (self.radius * math.cos(d * cstep), self.radius * math.sin(d * cstep))
      bcircle.append((v[0], v[1], self.height / 2.0))
      tcircle.append((v[0], v[1], -self.height / 2.0))

    # Origins Bottom      Top
    verts = [(0, 0, -self.height / 2.0), (0, 0, self.height / 2.0)]
    verts += (bcircle)
    verts += (tcircle)
    # Create circle faces from the individual circles.
    faces = []
    bcircle_range = np.arange(2, 2 + len(bcircle))
    tcircle_range = np.arange(2 + len(bcircle), 2 + len(bcircle) + len(tcircle))
    for i in range(self.divs):
      bf = (0, bcircle_range[i - 1], bcircle_range[i])
      tf = (1, tcircle_range[i - 1], tcircle_range[i])
      faces.append(bf)
      faces.append(tf)
    # Now that circle faces are complete, let's make the individual faces between them.
    # We can do this by iterating through both top & bottom circles 2 verts at a time
    # and creating a divided square.

    for i in range(self.divs):
      tl = tcircle_range[i - 1]
      tr = tcircle_range[i]
      bl = bcircle_range[i-1]
      br = bcircle_range[i]
      faces.append((tl, br, bl))
      faces.append((tl, tr, br))
    print(verts)
    print("")
    print(faces)

if __name__=="__main__":
  p = Pipe()



''' Tutorial Page '''

from PyQt5.QtWidgets import (
  QWidget,
  QGridLayout,
  QLabel,
  QPushButton,
  QVBoxLayout,
  QHBoxLayout,
  QStackedWidget,
  QShortcut)
from PyQt5.QtGui import QPixmap

class GettingStarted(QWidget):
  def __init__(self):
      super().__init__()
      self.page_number = 0
      self.layout_list = []
      self.init_ui()

  def init_ui(self):
    #sets window title, layout, and adds the widgets to the window
    self.setWindowTitle('L-System Help')
    self.main_layout = QVBoxLayout()
    self.button_layout = QHBoxLayout()
    self.pages_widget = QStackedWidget()
    self.pages_widget.setStyleSheet('''
          QLabel#title{
            font-weight: bold;
            font-size: 16px;
            text-decoration: underline;
            text-align: center;}

            QLabel{
                font-size: 14px;
                }
    ''')
    
    self.page_one_widget = QWidget() #axiom
    self.page_one_layout = QGridLayout()
    self.page_two_widget = QWidget() #production rules
    self.page_two_layout = QGridLayout()
    self.page_three_widget = QWidget() #Angle
    self.page_three_layout = QGridLayout()
    self.page_four_widget = QWidget() #iterations
    self.page_four_layout = QGridLayout()
    self.page_five_widget = QWidget() #branching
    self.page_five_layout = QGridLayout()
    self.page_six_widget = QWidget() #turning angle
    self.page_six_layout = QGridLayout()
    self.page_seven_widget = QWidget() #line scale
    self.page_seven_layout = QGridLayout()
    
    self.create_widgets()
    self.add_widgets()
    self.setLayout(self.main_layout)

  def create_widgets(self):
    # creates labels and buttons for the window
    self.axiom_title = QLabel('Axiom', objectName = 'title')
    self.axiom_one = QLabel('The axiom is where your L-System starts.')
    self.axiom_two = QLabel('Use production rules to transform your axiom into an L-System.')
    self.axiom_three = QLabel('See the symbols glossary for valid inputs.')
    self.axiom_screenshot = QLabel(self)
    self.axiom_pic = QPixmap("assets/screenshots/axiom.PNG");
    self.axiom_screenshot.setPixmap(self.axiom_pic)

    self.prod_rule_title = QLabel('Production Rules', objectName = 'title')
    self.prod_rule_one = QLabel('Production rules are part of how you transform your axiom into an L-System.')
    self.prod_rule_two = QLabel('This string will then be used to generate your L-System.')
    self.prod_rule_three = QLabel('Valid rules are in the form X:Y where X is a character or characters in your L-System ' +
                                  'and Y is which character(s) you want to replace it with.')
    self.prod_rule_four = QLabel('The box on the right side is for the weight of the rule.\n' +
                                 'If you only have one rule for a symbol than this box does nothing, ' +
                                 'if you have multiple rules with the same symbol on the left side ' +
                                 'you can weight them so one is chosen more often than the others.')
    self.prod_rule_five = QLabel('The sum of all boxes for a left side must add up to 1.')
    self.prod_rule_six = QLabel('The plus buttons on the right can be used to add up to 3 more production rules to your L-System')
    self.prod_rule_screenshot = QLabel(self)
    self.prod_rule_pic = QPixmap("assets/screenshots/prod_rules.PNG");
    self.prod_rule_screenshot.setPixmap(self.prod_rule_pic)
    
    self.angle_title = QLabel('Angle', objectName = 'title')
    self.angle_one = QLabel('The angle determines how steep the turns are in your L-System when you use + or -.')
    self.angle_two = QLabel('Valid inputs are any number from -360 to 360')
    self.angle_screenshot = QLabel(self)
    self.angle_pic = QPixmap("assets/screenshots/angles.PNG");
    self.angle_screenshot.setPixmap(self.angle_pic)

    self.iteration_title = QLabel('Iterations', objectName = 'title')
    self.iteration_one = QLabel('Iterations affect the number of times rules get applied to the string.')
    self.iteration_two = QLabel('The more iterations, the more defined the shape of your L-System becomes.')
    self.iteration_three = QLabel('Valid inputs are any number greater than 0.')
    self.iteration_screenshot = QLabel(self)
    self.iteration_pic = QPixmap("assets/screenshots/iterations.PNG");
    self.iteration_screenshot.setPixmap(self.iteration_pic)

    self.branching_title = QLabel('Branching', objectName = 'title')
    self.branching_one = QLabel('Branching allows for the L-System to create offshoots')
    self.branching_two = QLabel('You can add a branch by placing symbols between brackets')
    self.branching_three = QLabel('For example [F] will create a straight line branching from the main system')

    self.turn_angle_title = QLabel('Turning Angle', objectName = 'title')
    self.turn_angle_one = QLabel('The turning angle text box only appears if the axiom or a production rule has the symbols ( or ) in it.')
    self.turn_angle_two = QLabel('( decreases the angle by the turning angle and ) increases it.')
    self.turn_angle_three = QLabel('Valid inputs are any number between 0 and 360.')
    self.turn_angle_screenshot = QLabel(self)
    self.turn_angle_pic = QPixmap("assets/screenshots/turn_angle.PNG");
    self.turn_angle_screenshot.setPixmap(self.turn_angle_pic)

    self.line_scale_title = QLabel('Line Scaling', objectName = 'title')
    self.line_scale_one = QLabel('The line scaling text box only appears if the axiom or a production rule has the symbols < or > in it.')
    self.line_scale_two = QLabel('< divides the length of all subsequent lines by the line scaling factor and > increases it.')
    self.line_scale_three = QLabel('Valid inputs are any number')
    self.line_scale_screenshot = QLabel(self)
    self.line_scale_pic = QPixmap("assets/screenshots/line_scale.PNG");
    self.line_scale_screenshot.setPixmap(self.line_scale_pic)
    
    self.next_button = QPushButton("Next Page");
    self.next_button.setShortcut('Ctrl+N')
    self.next_button.clicked.connect(lambda: self.change_page(True))
    self.previous_button = QPushButton("Previous Page");
    self.previous_button.setShortcut('Ctrl+P')
    self.previous_button.clicked.connect(lambda: self.change_page(False))
    
  
  def add_widgets(self):
    # adds widgets to the window
    self.button_layout.addWidget(self.previous_button)
    self.button_layout.addStretch()
    self.button_layout.addWidget(self.next_button)

    self.page_one_layout.addWidget(self.axiom_title, 0, 0)
    self.page_one_layout.addWidget(self.axiom_one, 1, 0)
    self.page_one_layout.addWidget(self.axiom_two, 2, 0)
    self.page_one_layout.addWidget(self.axiom_three, 3, 0)
    self.page_one_layout.addWidget(self.axiom_screenshot,4,0)
    self.page_one_widget.setLayout(self.page_one_layout)
    self.layout_list.append(self.page_one_widget)
    self.pages_widget.addWidget(self.page_one_widget)

    self.page_two_layout.addWidget(self.prod_rule_title, 0, 0)
    self.page_two_layout.addWidget(self.prod_rule_one, 1, 0)
    self.page_two_layout.addWidget(self.prod_rule_two, 2, 0)
    self.page_two_layout.addWidget(self.prod_rule_three, 3, 0)
    self.page_two_layout.addWidget(self.prod_rule_four, 4, 0)
    self.page_two_layout.addWidget(self.prod_rule_five, 5, 0)
    self.page_two_layout.addWidget(self.prod_rule_six, 6, 0)
    self.page_two_layout.addWidget(self.prod_rule_screenshot, 7, 0)
    self.page_two_widget.setLayout(self.page_two_layout)
    self.layout_list.append(self.page_two_widget)
    self.pages_widget.addWidget(self.page_two_widget)

    self.page_three_layout.addWidget(self.angle_title, 0, 0)
    self.page_three_layout.addWidget(self.angle_one, 1, 0)
    self.page_three_layout.addWidget(self.angle_two, 2, 0)
    self.page_three_layout.addWidget(self.angle_screenshot, 3, 0)
    self.page_three_widget.setLayout(self.page_three_layout)
    self.layout_list.append(self.page_three_widget)
    self.pages_widget.addWidget(self.page_three_widget)

    self.page_four_layout.addWidget(self.iteration_title, 0, 0)
    self.page_four_layout.addWidget(self.iteration_one, 1, 0)
    self.page_four_layout.addWidget(self.iteration_two, 2, 0)
    self.page_four_layout.addWidget(self.iteration_three, 3, 0)
    self.page_four_layout.addWidget(self.iteration_screenshot, 4, 0)
    self.page_four_widget.setLayout(self.page_four_layout)
    self.layout_list.append(self.page_four_widget)
    self.pages_widget.addWidget(self.page_four_widget)

    self.page_five_layout.addWidget(self.branching_title, 0, 0)
    self.page_five_layout.addWidget(self.branching_one, 1, 0)
    self.page_five_layout.addWidget(self.branching_two, 2, 0)
    self.page_five_layout.addWidget(self.branching_three, 3, 0)
    self.page_five_widget.setLayout(self.page_five_layout)
    self.layout_list.append(self.page_five_widget)
    self.pages_widget.addWidget(self.page_five_widget)

    self.page_six_layout.addWidget(self.turn_angle_title, 0, 0)
    self.page_six_layout.addWidget(self.turn_angle_one, 1, 0)
    self.page_six_layout.addWidget(self.turn_angle_two, 2, 0)
    self.page_six_layout.addWidget(self.turn_angle_three, 3, 0)
    self.page_six_layout.addWidget(self.turn_angle_screenshot, 4, 0)
    self.page_six_widget.setLayout(self.page_six_layout)
    self.layout_list.append(self.page_six_widget)
    self.pages_widget.addWidget(self.page_six_widget)

    self.page_seven_layout.addWidget(self.line_scale_title, 0, 0)
    self.page_seven_layout.addWidget(self.line_scale_one, 1 ,0)
    self.page_seven_layout.addWidget(self.line_scale_two, 2, 0)
    self.page_seven_layout.addWidget(self.line_scale_three, 3, 0)
    self.page_seven_layout.addWidget(self.line_scale_screenshot, 4, 0)
    self.page_seven_widget.setLayout(self.page_seven_layout)
    self.layout_list.append(self.page_seven_layout)
    self.pages_widget.addWidget(self.page_seven_widget)

    self.main_layout.addWidget(self.pages_widget)
    self.main_layout.addStretch()
    self.main_layout.addLayout(self.button_layout)

  def change_page(self, up):
    if(up):
      self.page_number+=1
      self.page_number%=len(self.layout_list)
    else:
        self.page_number-=1
        if(self.page_number < 0):
          self.page_number+=len(self.layout_list)
          
    self.pages_widget.setCurrentIndex(self.page_number)



''' The glossary page '''
from PyQt5.QtWidgets import QWidget, QGridLayout, QLabel

class Glossary(QWidget):
    ''' Class of the glossary '''
    def __init__(self):
        ''' Initializes the glossary class with all of the labels  '''
        super().__init__()
        self.big_f = QLabel('F: draws a line of unit length.')
        self.little_f = QLabel('f: Moves forward in a line of unit length.')
        self.big_g = QLabel('G: draws a line of unit length.')
        self.little_g = QLabel('g: moves forward in a line of unit length.')
        self.big_h = QLabel('H: draws a line of half unit length.')
        self.little_h = QLabel(
            'h: moves forward in a line of half unit length.')
        self.plus = QLabel('+: turns the angle clockwise by an angle.')
        self.minus = QLabel('-: turns the angle counter-clockwise by an angle.')
        self.pipe = QLabel('|: reverses direction.')
        self.left_bracket = QLabel(
            '[: starts a branch. Must have a ] somewhere after.')
        self.right_bracket = QLabel(
            ']: ends a branch. Must have a [ somewhere before.')
        self.left_parenthesis = QLabel(
            '(: decreases the angle by a turning angle.')
        self.right_parenthesis = QLabel(
            '): increses the angle by a turning angle.')
        self.left_angle = QLabel(
            '<: divides line length by the length factor.')
        self.right_angle = QLabel(
            '>: multiplies the line length by the length factor.')
        self.control = QLabel(
            'A-E, I-Z: Control characters to control how the curve advances')
        self.init_ui()

    def init_ui(self):
        ''' Sets window title, layout, and adds widgets to the window '''
        self.setWindowTitle('Symbol Glossary')
        self.layout = QGridLayout()
        self.add_widgets()
        self.setLayout(self.layout)

    def add_widgets(self):
        ''' Adds the labels to the layout '''
        self.layout.addWidget(self.big_f, 1, 0)
        self.layout.addWidget(self.little_f, 2, 0)
        self.layout.addWidget(self.big_g, 3, 0)
        self.layout.addWidget(self.little_g, 4, 0)
        self.layout.addWidget(self.big_h, 5, 0)
        self.layout.addWidget(self.little_h, 6, 0)
        self.layout.addWidget(self.plus, 7, 0)
        self.layout.addWidget(self.minus, 8, 0)
        self.layout.addWidget(self.left_bracket, 9, 0)
        self.layout.addWidget(self.right_bracket, 10, 0)
        self.layout.addWidget(self.left_parenthesis, 11, 0)
        self.layout.addWidget(self.right_parenthesis, 12, 0)
        self.layout.addWidget(self.left_angle, 13, 0)
        self.layout.addWidget(self.right_angle, 14, 0)
        self.layout.addWidget(self.control, 15, 0)



# Predefined colors. 
import numpy as np

class Colors:
    """
    Convenience class pre-defining some colors to use with our mesh objects.

    Attributes:
    red (numpy.array): Color red.
    blue (numpy.array): Color blue.
    green (numpy.array): Color green.
    white (numpy.array): Color white.
    black (numpy.array): Color black.
    yellow (numpy.array): Color yellow.
    orange (numpy.array): Color orange.
    purple (numpy.array): Color purple.
    pink (numpy.array): Color pink.
    """
    red = np.array([1,0,0,0])
    blue = np.array([0,0,1,0])
    green = np.array([0,1,0,0])
    white = np.array([1,1,1,0])
    black = np.array([0,0,0,0])
    yellow = np.array([1,1,0,0])
    orange = np.array([1,0.5,0,0])
    purple = np.array([0.8,0,1,0])
    pink = np.array([1,0.8,0.945,0])



"""This file handles the OpenGL Window actions"""
from time import time
import math
# Other includes
import numpy as np
# Lsystem includes
from glm import vec3
# Python core includes
from PIL import Image
# PyQt5 includes
from PyQt5.QtWidgets import QApplication, QLabel, QVBoxLayout, QWidget

from lsystem.graphics.mesh import MeshObject

from pyqtgraph.opengl import GLLinePlotItem, GLMeshItem, GLViewWidget


class LSystem3DWidget(GLViewWidget):
    """
    This class manages the 3D pyqtgraph widget used to display the MeshObjects of the lsystem.

    Attributes:
    mesh (MeshObject): The combined mesh that's actually displayed.
    mesh_positions (list): A list tracking the positions of each mesh, used in calculating the position of the combined mesh.
    original_fov (vec3): Field of view at initialization, used for resetting the camera.
    original_distance (vec3): Distnace of the camera at initialization, used for resetting the camera.
    """
    def __init__(self):
        """
        Constructor of LSystem3DWidget class.
        """
        super(LSystem3DWidget, self).__init__()

        # Production scene objects.
        self.mesh = None
        self.mesh_positions = []
        self.original_fov = self.opts['fov']
        self.original_distance = self.opts['distance']

    def reset_camera(self):
        """
        Resets the camera to the initial position & zoom level.
        """
        self.reset_zoom()

    def reset_zoom(self):
        """
        Resets the zoom level of the camera.
        """
        self.opts['fov'] = self.original_fov
        self.opts['distance']=self.original_distance
        self.update()

    def screenshot(self, filename, pos):
        """
        Captures the view of the widget and saves it to a filename on disk.

        Parameters:
        filename (str): Filename to save to.
        pos (???): Nani the fuck is this shit?
        """
        #print("[ INFO ] Intentionally broken at the moment. There is a bug in pyqtgraph's image exporter so we'll need to fork or do it ourselves")
        #rect is the rectange (x,y,width,height) of the widget
        # rect = self.frameGeometry()
        # #shift over to get absl. pos
        # rect.translate(pos)
        # im = ImageGrab.grab(bbox=(rect.x(), rect.y(), rect.x()+rect.width(), rect.y()+rect.height())) # X1,Y1,X2,Y2
        # im.save(filename)
        self.grabFrameBuffer().save(filename)
        # self.renderPixmap().save(filename)

    def clear(self):
        """Removes all mesh objects from the scene."""
        self.mesh_positions.clear()
        if(self.mesh is not None):
            self.removeItem(self.mesh)


    def add_mesh(self, mesh_list: list):
        """
        Adds MeshObjects to the scene.

        Parameters:
        mesh_list (list): List of MeshObjects.
        """
        meshes = []
        for m in mesh_list:
          for mesh in m:
            self.mesh_positions.append(mesh.opts["position"])
            meshes.append(mesh)

        self.mesh = self.combine_meshes(meshes)
        self.addItem(self.mesh)
        self.center_mesh()


    def combine_meshes(self, mesh_list: list):
        """
        Combines all added MeshObjects to one singular mesh. This mimimizes OpenGL calls used underneath and makes working with the data easier.

        Parameters:
        mesh_list (list): List of MeshObjects

        Returns:
        MeshObject: Combination of the passed MeshObjects in mesh_list.
        """
        # So can't we do this by tracking the offset of the vert list when we add the vertexes
        # Then when we add the faces we just add the offset to each face?
        verts = np.array([])
        faces = np.array([])
        for mesh in mesh_list:
            offset = len(verts)
            verts = np.append(verts, mesh.opts["vertexes"])
            faces = np.append(faces, mesh.opts["indices"] + offset)
            verts = verts.reshape(math.floor(verts.shape[0]/3),3)
            faces = faces.reshape(math.floor(faces.shape[0]/3),3)
        faces = faces.astype(int)
        return MeshObject(vertexes=verts, indices=faces)

    def center_mesh(self):
      """
      Centers the mesh on the origin for a better camera experience.
      """
      positions = np.array(self.mesh_positions)
      avg_x = np.average(positions[:, 0])
      avg_y = np.average(positions[:, 1])
      avg_z = np.average(positions[:, 2])
      center = (avg_x, avg_y, avg_z)
      self.mesh.set_position(self.mesh.opts["position"] - center)



from copy import deepcopy
from math import degrees, pi, radians

import numpy as np
import pyqtgraph as pg
import pyqtgraph.opengl as gl
import PyQt5
from glm import vec3, vec4
from pyqtgraph.opengl import GLMeshItem, MeshData

from scipy.spatial.transform import Rotation as R
from lsystem.graphics.spatial_object import SpatialObject


class MeshObject(SpatialObject, GLMeshItem):
    """Interface to pyqtgraph's GLMeshItem that allows us to scale and rotate the vertices.
    Attributes:
      opts (dict): Dictionary containing majority of the class' variables. At one point this was necessary due to passing some kwargs to a parent class, but is now pointless.
            vertexes: Array of vertexes in world coordinates.
            model_verts: Array of vertexes in model coordinates
            indices: Indices of the faces.
            position: World position
            rot_vec: Current rotation in the vector form.
            rot_quat: Current rotation in quaternion form
            scale: Scale of the model
            smooth: I forget.
            computeNormals: Do we compute normals? Nah. No lighting no need.
            drawEdges: Do we draw edges?
            drawFaces: Do we draw faces?
            shader: Custom shader - Deprecated.
            color: Color of the mesh.
            edgeColor: Color of the edges of the mesh.
    """

    def __init__(self, vertexes=None, indices=None, *args, **kwds):
        """
        Constructor for the MeshObject class.

        Parameters:
        vertexes (list): List of vertices used in the mesh.
        indices (list): List of indices for the faces used in the mesh.
        rot_quat (glm.quat): Quaternion defining the rotation of the object.
        rot_vec (vec3): Vector3 defining the rotation of the object.
        scale (vec3): Scale factor for each axis of the object. Defaults to (1.0, 1.0, 1.0)
        position (vec3): Position of the object. Defaults to (0.0, 0.0, 0.0)
        """
        GLMeshItem.__init__(self)
        SpatialObject.__init__(self)
        # Argument extraction
        self.opts = {
            "vertexes": None,
            "model_verts": None,
            "indices": None,
            "position": vec3(0.0),
            "rot_vec": None,
            "rot_quat": None,
            "scale": vec3(1.0),
            "smooth": True,
            "computeNormals": False,
            "drawEdges": False,
            "drawFaces": True,
            "shader": None,
            "color": (1.0, 1.0, 1.0, 1.0),
            "edgeColor": (0.5, 0.5, 0.5, 1.0),
        }
        for k, v in kwds.items():
            self.opts[k] = v
        if(vertexes is not None):
            self.opts["vertexes"] = vertexes
        if(indices is not None):
            self.opts["indices"] = indices
            
        self.set_position(self.opts["position"], False)
        if (self.opts["rot_quat"] is not None):
          self.rotation = R.from_quat(self.opts["rot_quat"])
        if (self.opts["rot_vec"] is not None):
          self.rotation = R.from_rotvec(self.opts["rot_vec"])
        self.set_scale(self.opts["scale"], False)
        self.opts["meshdata"] = MeshData()
        self.set_vertexes(self.opts["vertexes"], False)
        self.set_indices(self.opts["indices"], False)
        if (self.opts["vertexes"] is not None):
            self.update_vertices()

    def get_vertices(self):
        """Returns an array of vertexes in world coordinates."""
        return self.opts["vertexes"]

    def get_faces(self):
        """Returns an array of face indices."""
        return self.opts["indices"]

    def __reduce__(self):
        """Necessary for pickling & deep copying"""

        return (MeshObject, (self.get_vertices(), self.get_faces()))

    def update_vertices(self):
        """Updates the world vertices with the current scale/translation/rotation."""
        model = self.generate_model_matrix()
        rotation = self.get_rotation()
        verts = deepcopy(self.opts["model_verts"])
        nverts = []
        for v in verts:
            nv = rotation.apply(v)
            nv = np.array(model * vec4(nv, 1.0))
            nv = nv[:3]
            nverts += [nv]

        self.opts["vertexes"] = np.array(nverts, dtype=np.float32)
        self.opts["meshdata"].setVertexes(self.opts["vertexes"])
        if self.opts["indices"] is not None:
            self.opts["meshdata"].setFaces(self.opts["indices"])
        self.setMeshData(meshdata=self.opts["meshdata"])

    def set_position(self, pos, update=True):
        """Sets the position and optionally triggers an update of the vertices.

        Parameters:
        pos (vec3): 3 dimensional vector representing the new position of the object.
        update (boolean): True/False for whether to update the graphics card. Useful for bulking together GPU changes instead of switching contexts a bunch.
        """
        super(MeshObject, self).set_position(pos)
        self.opts["position"] = pos
        if update:
            self.update_vertices()

    def set_rotation(self, o, update=True):
        """Sets the rotation and optionally triggers an update of the vertices.

        Parameters:
        o (vec3): Vector3 representing the new orientation/rotation of the object.
        update (boolean): True/False for whether to update the graphics card. Useful for bulking together GPU changes instead of switching contexts a bunch.
        """
        super(MeshObject, self).set_rotation(o)
        self.opts["rotation"] = o
        if update:
            self.update_vertices()

    def set_scale(self, s, update=True):
        """Sets the scale and optionally triggers an update of the vertices.

        Parameters:
        update (boolean): True/False for whether to update the graphics card. Useful for bulking together GPU changes instead of switching contexts a bunch.
        """
        super(MeshObject, self).set_scale(s)
        self.opts["scale"] = s
        if update:
            self.update_vertices()

    def set_vertexes(self, v, update=True):
        """Sets the array of vertexes and optionally triggers an update of the vertices.

        Parameters:
        v (list): List of vertices to use in the mesh.
        update (boolean): True/False for whether to update the graphics card. Useful for bulking together GPU changes instead of switching contexts a bunch.
        """
        if v is None:
            return
        # If these aren't floats, then there is an error in normal calculation
        self.opts["model_verts"] = np.array(v, dtype=np.float32)
        if update:
            self.update_vertices()

    def set_indices(self, i, update=True):
        """Sets the array of indices and optionally triggers an update of the vertices.

        Parameters:
        i (list): List of the vertex indices used to define the faces of the mesh.
        update (boolean): True/False for whether to update the graphics card. Useful for bulking together GPU changes instead of switching contexts a bunch.
        """
        if i is None:
            return
        self.opts["indices"] = np.array(i)
        if update:
            self.update_vertices()



"""This module makes the stack loop"""
import math
from time import time

import numpy as np


def read_substring(
    lsys, curr_state, turn_angle, trig_dict,  line_scale
):
    """
    This is a utility function for read_stack, takes in a string without
    branching ([, ]) or move without draw (f)

    Inputs:
        lsys: a string
        curr_state: a dictionary that has current point, orientation, and line scale
        turn_angle: the degrees the angle should increment by when seeing ( or )
        trig_dict: a dictionary of all the already calculated trig necassary for step direction
                   as well as the angle to turn by when seeing + or -
        line_scale: the factor by which the line length should change when seeing < or >
    Outputs1:
        An array of vertices formatted as such:  [[x1,y1],[x2,y2],[x3,y3],...]
    """
    vert_container = []  # make sure it's empty
    new_point = curr_state['point']  # initalize starting point
    new_angle = curr_state['angle']  # initalize starting angle
    vert_container.append(new_point)  # append first point
    for char in lsys:
        if not new_angle in trig_dict.keys():
            trig_dict[new_angle] = [
                math.cos(new_angle * np.pi / 180),
                math.sin(new_angle * np.pi / 180),
            ]
        if char == "F":
            new_point = [
                new_point[0] + (curr_state['scale'] * trig_dict[new_angle][0]),
                new_point[1] + (curr_state['scale'] * trig_dict[new_angle][1]),
                0,
            ]
            vert_container.append(new_point)
        elif char == "H":
            new_point = [
                new_point[0] + (scale * trig_dict[new_angle][0] / 2),
                new_point[1] + (scale * trig_dict[new_angle][1] / 2),
                0,
            ]
            vert_container.append(new_point)
        elif char == "+":
            new_angle = round((new_angle - trig_dict["angle"]) % 360, 5)
        elif char == "-":
            new_angle = round((new_angle + trig_dict["angle"]) % 360, 5)
        elif char == "|":
            new_angle = round((new_angle + 180) % 360, 5)
        elif char == "(":
            trig_dict["angle"] = round((trig_dict["angle"] - turn_angle) % 360, 5)
            # new_angle = round((new_angle - turn_angle)%360,5)
        elif char == ")":
            trig_dict["angle"] = round((trig_dict["angle"] + turn_angle) % 360, 5)
            # new_angle = round((new_angle + turn_angle)%360,5)
        elif char == ">":
            curr_state['scale'] *= curr_state['scale']
        elif char == "<":
            curr_state['scale']  /= curr_state['scale']

    # returns angle that string left off on, array of vertices, and the new angle
    return new_angle, vert_container, curr_state['scale']


def read_stack(stack, starting_pt, angle, turn_angle, line_scale):
    """
    This function converts a string into an array of objects placed at the vertices
    determined by the turtle graphics interpretation of the string

    Inputs:
        stack: a string
        starting_pt: the turtles starting_pt (the origin of your fractal)
        angle: the degrees the angle should turn when seeing a + or -
        turn_angle: the degrees the angle should increment by when seeing ( or )
        line_scale: the factor by which the line length should change when seeing < or >
    Outputs:
        An array of vertices formatted as such
        [
            [[x1,y1],[x2,y2],[x3,y3],...],
            [[x4,y4],[x5,y5],[x6,y6],...]
            ...
        ]
        where [x1,y1] is connected by a line to [x2,y2] which is connected by a line to [x3,y3]
        and [x4,y4] marks the starting point of a separte line
    """
    stack = stack.replace("G", "F")
    stack = stack.replace("g", "f")
    vertices = []
    vert_arr = []
    tmp_stack = []
    saved_states = []
    # keep the delimeter as the first character of the string
    while len(stack) > 0:
        index_start_b = stack[1:].find("[")  # index of staring bracket
        index_end_b = stack[1:].find("]")  # index of end bracket
        index_f = stack[1:].find("f")  # index of little f
        index_h = stack[1:].find("h")  # index of little h
        if max([index_start_b, index_end_b, index_f, index_h]) == -1:
            tmp_stack.append(stack)
            stack = []
        else:
            next_break = min(
                i for i in [index_start_b, index_end_b, index_f, index_h] if i >= 0
            )
            tmp_stack.append(stack[0 : next_break + 1])
            stack = stack[next_break + 1 :]
    # Set up a dictionary of all the possible angles
    # calculate the sin and cos of those angles ahead of time
    # WARNING currently rounding all angles to 5 digits, may not be exact enough

    # trig dict keeps track of the cos and sin of all the angles in radians
    # format trig_dict[an_angle] = [cos(an_angle), sin(an_angle)]
    # trig_dict["angle"] = the user inputted angle
    trig_dict = dict()
    trig_dict["angle"] = angle
    t = time()
    # new_point = starting_pt  # new point initalized to starting point
    curr_state = {"point": starting_pt, "angle": 0, "scale": float(1)}
    print(curr_state["angle"])
    print("TRIG DICT: ", trig_dict)
    # for each little f/h create a new array with the starting position and angle
    # initialized from the previous mesh
    for char in tmp_stack:
        if char[0] == "f" or char[0] == "h":
            if char[0] == "h":
                divisor = 2
                char.replace("h", "")
            else:
                divisor = 1
                char.replace("f", "")
            if curr_state["angle"] not in trig_dict.keys():
                trig_dict[curr_state["angle"]] = [
                    math.cos(curr_state["angle"] * np.pi / 180),
                    math.sin(curr_state["angle"] * np.pi / 180),
                ]
            curr_state["point"] = [
                curr_state["point"][0]
                + (curr_state["scale"] * trig_dict[curr_state["angle"]][0] / divisor),
                curr_state["point"][1]
                + (curr_state["scale"] * trig_dict[curr_state["angle"]][1] / divisor),
                0,
            ]
        elif char[0] == "[":
            saved_states.append(
                (curr_state["point"], curr_state["angle"], curr_state["scale"])
            )
            char.replace("[", "")
        elif char[0] == "]":
            tmp_state = saved_states.pop()
            curr_state["point"] = tmp_state[0]
            curr_state["angle"] = tmp_state[1]
            curr_state["scale"] = tmp_state[2]
            char.replace("]", "")

        curr_state["angle"], vertices, curr_state["scale"] = read_substring(
            char,
            curr_state,
            turn_angle,
            trig_dict,
            line_scale,
        )
        if len(vertices) != 1:
            vert_arr.append(vertices)
        #print("vert_arr = ",vert_arr)
        curr_state["point"] = vertices[-1]
    print("[ INFO ] Finshed finding vertices (", round(time() - t, 3), "s )")
    return vert_arr



'''This file handles the saving of rules'''
from PyQt5.QtWidgets import QWidget, QGridLayout, QLabel, QLineEdit, QPushButton
from lsystem.core.lsystem_utils import save_lsystem


class SaveRules(QWidget):
    '''Makes the window that allows the user to save rules'''
    def __init__(self, ui):
        '''Initing the name of the box and the ui'''
        super().__init__()
        self.ui = ui
        # creates the widgets to be added to the window
        self.save_label = QLabel("Name your rules")
        self.name_box = QLineEdit()
        self.name_box.returnPressed.connect(lambda: self.save())
        self.save_button = QPushButton("Save your rules")
        self.save_button.clicked.connect(lambda: self.save())

        self.init_ui()

    def init_ui(self):
        """sets the window title, layout, and adds widgets to the window"""
        self.setWindowTitle("Save your rules")
        self.layout = QGridLayout()
        self.add_widgets()
        self.setLayout(self.layout)

    def add_widgets(self):
        '''Adds the widgets to the layout'''
        # adds the widgets to the window
        self.layout.addWidget(self.save_label, 1, 0)
        self.layout.addWidget(self.name_box, 1, 1, 1, 2)
        self.layout.addWidget(self.save_button, 1, 3)

    def save(self):
        '''Saves the json file with the lsystem in it'''
        # grabs the name of the lsystem from the namebox, grabs the grammar from the ui,
        # and calls save_lsystem in lsystem_utils.py
        name = self.name_box.text()
        if len(name) > 0:
            grammar = {}
            grammar["rules"] = {}
            i = 0
            for rule in self.ui.prod_rules_edit:
                print(rule.text())
                rule = rule.text().replace(" ", "")
                prod = rule.split(":")
                # add the rule and probabilty to pr[0]'s list, if there isn't a list then make one
                try:
                    grammar["rules"][prod[0]].append(
                        [prod[1], self.ui.prod_percent[i].text()]
                    )
                except KeyError:
                    grammar["rules"][prod[0]] = []
                    grammar["rules"][prod[0]].append(
                        [prod[1], self.ui.prod_percent[i].text()]
                    )
                i += 1  # increment to grab the probability for each rule
            grammar["angle"] = float(self.ui.angle_edit.text())
            if self.ui.made_angle:
                grammar["turn_angle"] = float(self.ui.turn_angle_edit.text())
            else:
                grammar["turn_angle"] = 0
            if self.ui.made_line:
                grammar["line_scale"] = float(self.ui.line_scale_edit.text())
            else:
                grammar["line_scale"] = 0
            grammar["axiom"] = self.ui.axiom_edit.text()
            grammar["iterations"] = int(self.ui.iters_edit.text())
            save_lsystem(name, grammar)
            print("[ INFO ] L-System " + str(name) + " saved to disk...")
            # make window disappear
            self.hide()
        else:
            self.name_box.setStyleSheet("color: red;")
            self.name_box.setText("Name your L-System before you save!")



from lsystem.core.getting_started import GettingStarted
from lsystem.core.glossary import Glossary
from lsystem.core.my_ui import UIWidget
from lsystem.core.save_rules_window import SaveRules
from lsystem.core.settings import PopupSettings



""" This file for now is acting as a catch-all utility file."""
import json
import os
import platform
import copy
import logging
from pathlib import Path
import numpy as np
from lsystem.core.parsing import parsed_thread
from lsystem.core.stack_loop import read_stack as read_2d_stack
from lsystem.core.stack_loop3D import read_stack as read_3d_stack
from lsystem.core.graph import Graph
from lsystem.graphics.square import Square
from lsystem.graphics.pipe import Pipe

# At the moment the most important definitions are
# saved_lsystems - Global dictionary definition of loaded lsystems.
# save_lsystem - Save an lsystem to the user defined json file.
# load_saved_lsystems() - Loads the saved & predefined lsystems from file into grammar dictionary.
# get_saved_lsystem(key) - Returns a loaded lsystem from the dictionary.
# generate_lsystem(grammar) - Generates the vertices of a given lsystem
# normalize_coordinates(verts) - Normalizes verts given a numpy array.


# Global dictionary of lsystems.
# The dictionary uses the following keys.
# 'axiom' : string for the axiom
# 'rules' : dict of production rules
# 'angle' : the angle in degrees(integer)
# 'iterations' : number of iterations(integer)
# predefined areas to keep the files.


def get_saved_lsystem(key, saved_lsystems):
    """
    Fetches a saved lsystem in our loaded lsystems by key.

    Parameters:
    key (string): Dictionary key used to identify a saved lsystem.
    saved_lsystems (dict): Dictionary of loaded lsystems.

    Returns:
    dict: dictionary of the saved lsystem grammar, or None if it doesn't exist.
    """
    if key in saved_lsystems:
        grammar = saved_lsystems[key]
        return grammar
    else:
        print("[ ERROR ] No L-System loaded with key: " + str(key))
        return None


def generate_lsystem_2d(grammar):
    """
    Generates a 2D L-System based off of the grammar.

    Parameters:
    grammar (dict): Dictionary representing the grammar defining the lsystem.

    Returns:
    Graph: Graph object used to generate the mesh.
    """
    grammar_copy = copy.deepcopy(grammar)
    graph = Graph()  # Adjacency list based graph.
    print("[ INFO ] Generating L-System with the given grammar..." + str(grammar))
    # Generate full production string.
    tmp_stack = parsed_thread(
        grammar_copy["axiom"], grammar_copy["rules"], grammar_copy["iterations"]
    )
    # Generate vertics
    verts_arr_temp = read_2d_stack(
        tmp_stack,
        [0, 0, 0],
        grammar_copy["angle"],
        grammar_copy["turnAngle"],
        grammar_copy["lineScale"],
    )
    verts_arr_temp = np.array(verts_arr_temp)
    # for now manually strip the x and y values from verts_arr_temp
    x_vals = []
    y_vals = []
    for mesh in verts_arr_temp:
        for point in mesh:
            x_vals.append(point[0])
            y_vals.append(point[1])
    minx = min(x_vals)
    maxx = max(x_vals)
    miny = min(y_vals)
    maxy = max(y_vals)
    maxdif = max([maxx - minx, maxy - miny])
    # Find the min and max of x and y.
    # Find the max of abs(x_max-x_min) and abs(y_max-y_min)
    # Then perform (x-x_min)/max diff * .9999 and (y-y_min)/max diff * .9999

    # Edited for temporary support of adjacency lists. It's wasteful currently, but when we integrate it further back into the tech stack during refactor, it
    # should become more efficient.
    # Since each verts_arr_temp/verts represents an individual fork, we can assume they are connected by edges. So let's track the previous point and draw lines.
    for verts in verts_arr_temp:
        verts = np.array(verts, dtype=np.float32)
        verts = verts.reshape(verts.shape[0] * verts.shape[1])
        for i in range(0, len(verts), 3):  # change this to a 2 to make it 2D
            verts[i] = ((verts[i] - minx) * 0.99999) / maxdif
            verts[i + 1] = ((verts[i + 1] - miny) * 0.99999) / maxdif
            graph.add_vertex((verts[i], verts[i + 1]))
            if i > 0:
                graph.add_edge((prev_point), (verts[i], verts[i + 1]))
            prev_point = (verts[i], verts[i + 1])

    return graph

def generate_lsystem_3d(grammar):
    """
    Generates a 3D L-System based off of the grammar.

    Parameters:
    grammar (dict): Dictionary representing the grammar defining the lsystem.
    """
    grammar_copy = copy.deepcopy(grammar)
    graph = Graph()  # Adjacency list based graph.
    print("[ INFO ] Generating L-System with the given grammar..." + str(grammar))
    # Generate full production string.
    tmp_stack = parsed_thread(
        grammar_copy["axiom"], grammar_copy["rules"], grammar_copy["iterations"]
    )
    # Generate vertics
    mesh = read_3d_stack(
        tmp_stack,
        [0, 0, 0],
        grammar_copy["angle"],
        Pipe
    )

    return mesh

def get_saved_path():
  """Gets the operating system dependent saved_lsystems.json path

  Returns:
  - os.path: Operating system dependent saved_lsystems.json path
  """
  opsys = platform.system()
  if (opsys == 'Linux' or opsys == 'Darwin'):
    path = os.path.join(str(Path.home()), ".config/lsystem")
  elif(opsys=='Windows'):
    path = os.path.expandvars(r'%APPDATA%\lsystem')
  assert path is not None, "Something went wrong parsing operating system platform"
  try:
    os.makedirs(path)
  except FileExistsError:
    pass
  path = os.path.join(path, "saved_lsystems.json")
  return path


def save_lsystem(key, grammar):
    """
    Saves a given lsystem to disk to "lsystem/saved_lsystems.json"
    Overwrites any previous lsystem defined with the same key.

    Parameters:
    key (string): Key used to identify the lsystem.
    grammar (dict): Dictionary of the grammar defining the lsystem.

    Returns:
    - dict: New dictionary of lsystems with the new addition.
    """
    saved_lsystems = {}
    saved_file = get_saved_path()
    # Check if the file exists.
    if os.path.exists(saved_file):
        # If it does, then load all saved data and replace/insert the new data to the dict.
        with open(saved_file, "r") as sfile:
            saved = json.load(sfile)
            saved[key] = grammar
    else:
        # if it doesn't exist, we just create a new dict
        saved = {key: grammar}
    # Then overwrite the file.
    with open(saved_file, "w") as sfile:
        json.dump(saved, sfile, indent=2)
    return saved_lsystems

def load_saved_lsystems():
    """
    Loads the pre-defined and user defined L'Systems.

    Returns:
    - dict: Dictionary of all saved lsystems in the form of {key:grammar, key:grammar}
    """
    saved_lsystems = {}
    saved_file = get_saved_path()
    predef_file = os.path.join(os.path.dirname(__file__), "../assets/lsystems/predefined_lsystems.json")
    print(predef_file)
    print("[ INFO ] Loading saved L-Systems from disk...")
    # Check if the file exists.
    if os.path.exists(predef_file):
        # If it does, then load it as a json object.
        predef = json.load(open(predef_file, "r"))
        # For every key(aka lsystem definition), add it to our saved lsystems.
        for key in predef.keys():
            saved_lsystems[key] = predef[key]

    # Check if the file exists.
    if os.path.exists(saved_file):
        # If it does, then load it as a json object.
        saved = json.load(open(saved_file, "r"))
        # For every key(aka lsystem definition), add it to our saved lsystems.
        for key in saved.keys():
            saved_lsystems[key] = saved[key]
    return saved_lsystems


def remove_saved_lsystem(key):
    """
    Deletes saved lsystem by key by loading the file into a json object, removing the key, then writing the file back to disk.

    Parameters:
    key (string): Key of the lsystem to remove.
    """
    saved_file = "assets/lsystems/saved_lsystems.json"
    # Check if the file exists.
    if os.path.exists(saved_file):
        # If it does, then load all saved data and delete the data, then resave it.
        with open(saved_file, "r") as sfile:
            saved = json.load(sfile)
            try:
                del saved[key]
            except KeyError:
                print("[ ERROR ] Key " + str(key) + " not found in saved lsystems")
    else:
        # if it doesn't exist, we just return
        return
    # Then overwrite the file.
    with open(saved_file, "w") as sfile:
        json.dump(saved, sfile, indent=2)



import numpy as np
from glm import vec3

from lsystem.graphics.mesh import MeshObject


class Square(MeshObject):
    """afjaksdflhas"""

    def __init__(self, pos=vec3(0.0), rotation=vec3(0.0)):
        MeshObject.__init__(self, position=pos, rotation=rotation)
        self.set_vertexes(np.array([
          (0.0, 0.0, 0.0),
          (0.0, 1.0, 0.0),
          (1.0, 0.0, 0.0),
          (1.0, 1.0, 0.0)
          ]), False)
        self.set_indices(np.array([[0, 1, 2], [1, 2, 3]]))
        

if __name__ == "__main__":
    m = Square()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="L-System-Visualizer",
    version="1.0.1",
    author="Team BRAWT",
    author_email="lbones1@gulls.salisbury.edu",
    keywords="pyside2 qt5 lsystem",
    description="Generates visualizations of L Systems after a certain number of productions.",
    url="https://github.com/wrathofrathma/L-System-Visualizer.git",
    license="Unknown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    install_requires=["PyQt5", "matplotlib", "numpy", "Pillow", "PyGLM", "PyOpenGL", "pyparsing", "pyqtgraph", "scipy", "pyscreenshot", "pyrr", "pywavefront", ],
    py_modules=['lsystem'],
    python_requires='>=3.7',
    package_data={
      "": ["*.json", "*.png", "*.jpg", "*.jpeg"]
      },
    include_package_data=True
)



"""This module makes the stack loop"""
import math
from time import time

import numpy as np
from scipy.spatial.transform import Rotation as R
def read_substring(
    lsys, curr_state, turn_angle, obj
):
    """
    This is a utility function for read_stack, takes in a string without
    branching ([, ]) or move without draw (f)

    Inputs:
        lsys: a string
        curr_state: a dictionary that has current point, and orientation
        turn_angle: the degrees the angle should turn when seeing a +, -, ^, &, < or >
        obj: the type of object that should be displayed, needs to be a mesh object
    Outputs:
        An array of type obj
    """
    obj_container = []  # make sure it's empty
    current_point = curr_state['point']  # initalize starting point
    #Rotation matrix converts a vector in the abs frame to turtle frame
    vertices = []
    orientation_mat= curr_state['orientation_mat']
    rotation=obj_ortientation = R.from_matrix(orientation_mat)
    vertices.append(current_point)
    #new_obj = obj(pos=current_point,rotation=obj_ortientation.as_rotvec()) # append first object
    #obj_container.append(new_obj)
    '''First row of the rotation matirx is the turtles forward
        direction vector in the absolute frame
        Second row of the rotation matrix is the turtles right
        direction vector in the absolute frame
        Third row of the rotation matrix is the turtles down
        direction vector in the absolute frame
    '''
    unit_step = curr_state['orientation_mat'][0]
    for char in lsys:
        if char == "F":
            old_point = current_point
            current_point = np.add(current_point, unit_step)
            vertices.append(current_point)
            obj_ortientation = R.from_matrix(orientation_mat)
            new_obj = obj(start = old_point, end = current_point, rotation=obj_ortientation.as_rotvec())
            obj_container.append(new_obj)
        elif char == "+":
            #turning left (about the turtle's down axis)
            rotation_vector = np.array(orientation_mat[2])*turn_angle
            rotation=R.from_rotvec(rotation_vector)
            orientation_mat = rotation.apply(orientation_mat)
            unit_step = orientation_mat[0]
        elif char == "-":
            #turning right (about the turtle's down axis)
            rotation_vector = np.array(orientation_mat[2])*turn_angle*(-1)
            rotation=R.from_rotvec(rotation_vector)
            orientation_mat = rotation.apply(orientation_mat)
            unit_step = orientation_mat[0]
        elif char == "<":
            #rolling left (about the turtle's forward axis)
            rotation_vector = np.array(orientation_mat[0])*turn_angle
            rotation=R.from_rotvec(rotation_vector)
            orientation_mat = rotation.apply(orientation_mat)
            unit_step = orientation_mat[0]
        elif char == ">":
            #rolling right (about the turtle's forward axis)
            rotation_vector = np.array(orientation_mat[0])*turn_angle*(-1)
            rotation=R.from_rotvec(rotation_vector)
            orientation_mat = rotation.apply(orientation_mat)
            unit_step = orientation_mat[0]
        elif char == "^":
            #pitch up (about the turtle's right axis)
            rotation_vector = np.array(orientation_mat[1])*turn_angle
            rotation=R.from_rotvec(rotation_vector)
            orientation_mat = rotation.apply(orientation_mat)
            unit_step = orientation_mat[0]
        elif char == "&":
            #pitch down (about the turtle's right axis)
            rotation_vector = np.array(orientation_mat[1])*turn_angle*(-1)
            rotation=R.from_rotvec(rotation_vector)
            orientation_mat = rotation.apply(orientation_mat)
            unit_step = orientation_mat[0]
    curr_state['point'] = current_point
    # returns angle that string left off on, array of vertices, and the new angle
    return orientation_mat, obj_container, vertices


def read_stack(stack, starting_pt, angle, obj):
    """
    This function converts a string into an array of objects placed at the vertices
    determined by the turtle graphics interpretation of the string

    Inputs:
        stack: a string
        starting_pt: the turtles starting_pt (the origin of your fractal)
        angle: the degrees the angle should turn when seeing a +, -, ^, &, < or >
        obj: the type of object that should be displayed, needs to be a mesh object
    Outputs:
        An array of type obj
    """
    angle = angle*np.pi/180
    stack = stack.replace("G", "F")
    stack = stack.replace("g", "f")
    array_of_objects = []
    twoD_array_objects = []
    tmp_stack = []
    saved_states = []
    # keep the delimeter as the first character of the string
    while len(stack) > 0:
        index_start_b = stack[1:].find("[")  # index of staring bracket
        index_end_b = stack[1:].find("]")  # index of end bracket
        index_f = stack[1:].find("f")  # index of little f
        index_h = stack[1:].find("h")  # index of little h
        if max([index_start_b, index_end_b, index_f, index_h]) == -1:
            tmp_stack.append(stack)
            stack = []
        else:
            next_break = min(
                i for i in [index_start_b, index_end_b, index_f, index_h] if i >= 0
            )
            tmp_stack.append(stack[0 : next_break + 1])
            stack = stack[next_break + 1 :]
    t = time()
    curr_state = {"point": starting_pt, "orientation_mat": np.array([[1, 0, 0],[0,1,0],[0,0,1]])}
    # for each little f/h create a new array with the starting position and angle
    # initialized from the previous mesh
    for char in tmp_stack:
        if char[0] == "f":
            char.replace("f", "")
            unit_step = curr_state['orientation_mat'][0]
            prev_point = curr_state['point']
            curr_state['point'] = np.add(curr_state['point'], unit_step)
            rotation=obj_ortientation = R.from_matrix(curr_state['orientation_mat'])
            new_obj = obj(start=prev_point, end=curr_state['point'], rotation=obj_ortientation.as_rotvec())
        elif char[0] == "[":
            saved_states.append(
                (curr_state["point"], curr_state["orientation_mat"])
            )
            char.replace("[", "")
        elif char[0] == "]":
            tmp_state = saved_states.pop()
            curr_state["point"] = tmp_state[0]
            curr_state["orientation_mat"] = tmp_state[1]
            char.replace("]", "")

        curr_state["orientation_mat"], array_of_objects, verts = read_substring(
            char,
            curr_state,
            angle,
            obj,
        )
        #if not(len(array_of_objects) == 1):
        twoD_array_objects.append(array_of_objects)
        #print(curr_state['point'])
        #curr_state["point"] = verts[-1]
    print("[ INFO ] Finshed finding vertices (", round(time() - t, 3), "s )")
    return twoD_array_objects

# if __name__ == "__main__":
#     str = "FF+F-F{F}"
#     starting_pt=[0,0,0]
#     curr_state = {"point": starting_pt, "angle": np.array([1,0,0], dtype=float), "scale": float(1)}
#     print(read_substring(str, curr_state, np.radians(45)))



'''This file handles the saving of rules'''
from PyQt5.QtWidgets import QWidget, QGridLayout, QLabel, QLineEdit, QPushButton
import numpy as np
import copy

class FractalDimension(QWidget):
    '''Makes the window that allows the user to save rules'''
    def __init__(self, ui):
        '''Initing the name of the box and the ui'''
        super().__init__()
        self.ui = ui
        # creates the widgets to be added to the window
        self.instructions = QLabel("Welcome to the fractal dimension calculator. We calculate based on powers of 2.")
        self.start_prompt = QLabel("Start Power: ")
        self.start_size = QLineEdit()
        self.end_prompt = QLabel("End Power: ")
        self.end_size = QLineEdit()
        self.iters_prompt = QLabel("Number of calculations")
        self.iters_num = QLineEdit()
        self.calc_button = QPushButton("Calculate")
        self.calc_button.clicked.connect(lambda: self.calc())

        self.init_ui()

    def init_ui(self):
        """sets the window title, layout, and adds widgets to the window"""
        self.setWindowTitle("Fractal Dimmension Calculator")
        self.layout = QGridLayout()
        self.add_widgets()
        self.setLayout(self.layout)

    def add_widgets(self):
        '''Adds the widgets to the layout'''
        # adds the widgets to the window
        self.layout.addWidget(self.instructions, 0, 0, 1, 5)
        self.layout.addWidget(self.start_prompt, 2, 0)
        self.layout.addWidget(self.start_size, 2, 1, 1, 2)
        self.layout.addWidget(self.end_prompt, 3, 0)
        self.layout.addWidget(self.end_size, 3, 1, 1, 2)
        self.layout.addWidget(self.iters_prompt, 4, 0)
        self.layout.addWidget(self.iters_num, 4, 1, 1, 2)
        self.layout.addWidget(self.calc_button, 5, 3)

    def calc(self):
      all_dims = []
      for i in range(int(self.iters_num.text())):
        self.ui.gen_sys()
        verts = copy.deepcopy(self.ui.verts)
        fractal_dim = []
        fract_avg = []
        x_arr = []
        y_arr = []

        start = 2 ** int(self.start_size.text())
        end = 2 ** int(self.end_size.text())

        print("Lela's code", verts)
        print("Start size = {}".format(self.start_size.text()))
        print("End size = {}".format(self.end_size.text()))

        pix_map = gen_pixel_map(verts, end)
        fractal_dim.insert(0, np.log2(np.count_nonzero(pix_map == '1')))
        while(end >= start):
            print("LELAS FRACT CALC END STATE", end)
            end = end/2
            pix_map = pool_pixel_map(pix_map)

            fractal_dim.insert(0, np.log2(np.count_nonzero(pix_map == '1')))


        for i, dim in enumerate(fractal_dim):
          x_arr.append(np.log2(start))
          y_arr.append(dim)
          start = start*2
        all_dims.append(np.polyfit(x_arr, y_arr, 1)[0])
      print(all_dims)
          


def gen_pixel_map(graph, size):
  #meshes is an array of arrays of vertices
  #add one and multiply by size to get from world coord to pixel map
  pix_map = np.full((size,size),'.')

  for point in graph.adjacency_list.keys():
    #second coordinate pair
    #print("point = ",point)
    x_old = point[0]*size #coordinates are stored as real numbers for precision
    y_old = point[1]*size
    pix_map[int(np.trunc(x_old)), int(np.trunc(y_old))]=1
    for point2 in graph.adjacency_list[point]["edges"]:
      #print("point 2 = ",point2)
      x_new = point2[0]*size
      y_new = point2[1]*size

      #add the line connection new point and old point
      h=0
      step_size=1/(max(abs(x_new-x_old), abs(y_new-y_old))*2)
      for h in np.arange(0,1,step_size):
        xx = x_old + h*(x_new-x_old)
        yy = y_old + h*(y_new-y_old)
        xx_coord = min(int(np.trunc(xx)),size-1)
        yy_coord = min(int(np.trunc(yy)),size-1)
        pix_map[xx_coord, yy_coord]=1
      #This takes care of edge case where max =1024 is outside of array
      x_new_coord = min(int(np.trunc(x_new)), size-1)
      y_new_coord = min(int(np.trunc(y_new)),size-1)

      pix_map[x_new_coord, y_new_coord]=1
    #x_old = x_new
    #y_old = y_new

  #print("pixel map done")
  return pix_map
  #np.set_printoptions(threshold=np.inf)
  #print(pix_map)
def pool_pixel_map(map):
  size = len(map)
  pix_map = np.full((int(size/2),int(size/2)),'.')
  for i in range(0,size,2):
    for j in range(0,size,2):
      if (map[i,j] == '1' or map[i+1,j]=='1' or map[i,j+1]=='1' or map[i+1,j+1]=='1'):
        pix_map[int(i/2), int(j/2)]=1
  return pix_map
def fractal_dim_calc(graph,ending_size, num_sizes):

  fractal_dim = []
  pix_map = gen_pixel_map(graph,ending_size)
  np.set_printoptions(threshold=np.inf, linewidth = 2000)


  count= 0
  fractal_dim.insert(0,np.log2(np.count_nonzero(pix_map == '1')))#/np.log(ending_size))
  for i in range(num_sizes-1):
    print("STEPHS FRACT_CALC", ending_size)
    ending_size = ending_size/2
    
    pix_map = pool_pixel_map(pix_map)
    #if >2:
    #  print(pix_map)
    fractal_dim.insert(0,np.log2(np.count_nonzero(pix_map == '1')))#/np.log(ending_size))

  return fractal_dim # JUST RETURNS log(N(E))




"""This file handles the popup settings window"""
from PyQt5.QtWidgets import QWidget, QGridLayout, QLabel, QButtonGroup, QRadioButton



class PopupSettings(QWidget):
    """This is the window for the popup settings"""

    def __init__(self, graphix):
        """Initializes important variables for the settings window"""
        super().__init__()
        self.graphix = graphix
        self.dimension = QLabel("Dimensions: ")
        self.two_dim = QRadioButton("2D")
        self.three_dim = QRadioButton("3D")
        self.init_UI()

    def init_UI(self):
        """Makes the UI for the settings menu"""
        self.setWindowTitle("L-systems Settings")
        self.layout = QGridLayout()
        self.init_buttons()
        self.add_widgets()
        self.setLayout(self.layout)

    def init_buttons(self):
        """Initializes the buttons and button states in the settings menu"""
        dim_group = QButtonGroup(self)

        self.two_dim.setChecked(True)
        self.two_dim.toggled.connect(lambda: self.dim_state(self.two_dim))
        dim_group.addButton(self.two_dim)
        dim_group.addButton(self.three_dim)

    def add_widgets(self):
        """Adds the widgets to the settings layout"""
        self.layout.addWidget(self.dimension, 1, 0)
        self.layout.addWidget(self.two_dim, 1, 1)
        self.layout.addWidget(self.three_dim, 1, 2)

    def dim_state(self, dim):
        """Sets the dimension state when changed"""
        if dim.text() == "2D" and dim.isChecked():
            pass
        else:
            pass



import tkinter as tk
import time

""" Author(s): Lela Bones and Stephanie Warman
    file: UI-Layout.py """

fields = ("Axiom", "Production Rules", "Angle (degrees)", "Iterations")
alphabet = ("F", "f", "+", "-")
err_mess = "invalid"
flash_delay = 500  # msec between colour change
flash_colours = ("black", "red")  # Two colours to swap between


def flashColor(object, color_index, init_time):
    if time.time() - init_time < 2:
        object.config(foreground=flash_colours[color_index])
        root.after(flash_delay, flashColor, object, 1 - color_index, init_time)
    else:
        object.config(foreground=flash_colours[0])


def clear_box(object, init_time):
    if time.time() - init_time < 2:
        root.after(flash_delay, clear_box, object, init_time)
    else:
        object.delete(0, tk.END)


def print_invalid(object):
    object.delete(0, tk.END)
    object.insert(0, err_mess)
    flashColor(object, 0, time.time())
    clear_box(object, time.time())


def get_Axiom(entries):
    """
  This function input checks the input for axiom
  Input: entries is an array collected from makeform
  Ouput: invalid if not valid input
  """
    axiom = entries["Axiom"].get()
    for ch in axiom:
        if not ch in alphabet:
            print_invalid(entries["Axiom"])

        # root.after(10, entries['Axiom'].config(background=dfault_col))


def get_ProdRules(entries):
    """
  This function input checks the input for production rules
  Input: entries is an array collected from makeform
  Ouput: invalid if not valid input
  """
    prodRule = entries["Production Rules"].get()
    prodRule = prodRule.replace(" ", "")
    if not "->" in prodRule or prodRule[1] == ">" or prodRule[len(prodRule) - 1] == ">":
        print_invalid(entries["Production Rules"])
    tmp_prodRule = prodRule.replace("->", "")
    for ch in tmp_prodRule:
        if not ch in alphabet:
            print_invalid(entries["Production Rules"])


def get_Angle(entries):
    """
  This function input checks the input for angle
  Input: entries is an array collected from makeform
  Ouput: invalid if not valid input
  """
    try:
        angle = float(entries["Angle (degrees)"].get())
    except:
        print_invalid(entries["Angle (degrees)"])
        return
    if angle <= -360 or angle >= 360:
        print_invalid(entries["Angle (degrees)"])


def get_Iter(entries):
    """
  This function input checks the input for iterations
  Input: entries is an array collected from makeform
  Ouput: invalid if not valid input
  """
    try:
        iterat = int(entries["Iterations"].get())
    except:
        print_invalid(entries["Iterations"])
        return
    if iterat <= 0:
        print_invalid(entries["Iterations"])


def check_inputs(entries):
    get_Axiom(entries)
    get_ProdRules(entries)
    get_Angle(entries)
    get_Iter(entries)


def makeform(root, fields):
    """ 
  This function makes the input for the application
  Input: root is the object of the tk window
         fields are the input fields
  Output: entries for the field
  """

    entries = {}
    root.title("L-System")
    for field in fields:
        row = tk.Frame(root)
        label = tk.Label(row, width=22, text=field + ": ", anchor="w")
        entry = tk.Entry(row)
        entry.insert(0, "")
        row.pack(side=tk.TOP, fill=tk.X, padx=5, pady=5)
        label.pack(side=tk.LEFT)
        entry.pack(side=tk.RIGHT, expand=tk.YES, fill=tk.X)
        entries[field] = entry

    return entries


if __name__ == "__main__":
    root = tk.Tk()
    entries = makeform(root, fields)
    frame = tk.Frame(width=200, height=200)
    frame.pack()
    b1 = tk.Button(
        root, text="Generate L-System", command=(lambda e=entries: check_inputs(e))
    )
    b1.pack(side=tk.LEFT)
    b2 = tk.Button(root, text="Quit", command=root.quit)
    b2.pack(side=tk.RIGHT)
    root.mainloop()
